"""
Path     addons/source-python/packages/custom/k2tools/cmdlib.py
Name     Command Library
Version  1.00.000
Revision $Rev: 39 $
Author   (C) [#OMEGA] - K2


INFO

NOTES

Discussion Thread here:
http://www.sourcepython.com/forums/showthread.php?93


ON SPAM PROTECTION

I found it difficult to provide an optimal solution to command spamming. Some
commands may be rather light and can be executed a lot while others can impact
the game greatly.

I originally intended per-command spam-protection, but that leads to the issue
that if /multiple/ commands are spammed people can bypass the protection.

The downside of using a shared spam-protection is that that we can't define it
as correctly as before. So currently, we have a global setting for the time
between execution attempts and a shared value for any command execution
attempts.
So, commands with a high execution limit threshold, can still be executed,
however ones with a lot execution limit may be not.


COMMAND EXECUTION METHOD RESOLUTION ORDER

I put the methods for the command execution in a way I think it makes the most
sense. In order to be able to modify this in a simple manner, each 'aspect' is
a method if it does a more complex task.


AGREEMENT

See k2tools_license.txt


TODO

Fix minor TODO-S
Eventually add much more detailed error-checking for command registry
Comments & doc-strings
  
"""

__k2doc__ = {'AdvancedCommandManager': 'instance'}
__all__ = ['AdvancedCommandManager']
__version__ = '1.00.000 $Rev: 39 $'
__author__ = '[#OMEGA] - K2'
__date__ = '$Date: 2014-02-09 22:15:20 +0100 (So, 09 Feb 2014) $'

# =============================================================================
# >> IMPORTS
# =============================================================================

# Python
import logging
import re
from time import time

# Source.Python - C
import command_c
import cvar_c
from conversions_c import index_from_playerinfo
from conversions_c import index_from_userid
from conversions_c import playerinfo_from_userid

# Source.Python - PY
from auth.manager import AuthManager
from core import GameEngine as _Engine
from events.manager import EventRegistry as _EventRegistry
from messages import SayText2

# K2 Tools

from ._shared import lang

# =============================================================================
# >> GLOBALS
# =============================================================================

_saycmd_re = re.compile(r'("[^"]+")|([^ :]+)|([:]+)')
_logger = logging.getLogger('k2tools.cmdlib')

# =============================================================================
# >> METHODS
# =============================================================================


def _access_denied(av):
    index = index_from_playerinfo(av['playerinfo'])
    l = _Engine.get_client_convar_value(index, 'cl_language')
    SayText2(lang['cmdlib_access_denied'].get_string(l), index).send(index)
    

def _kick(av):
    if av['playerinfo']:
        l = _Engine.get_client_convar_value(
            index_from_playerinfo(av['playerinfo']), 'cl_language')
        _Engine.server_command(
            'kickid %s "%s";' % (av['playerinfo'].get_userid(), 
            lang['cmdlib_kicked_for_command_spamming'].get_string(l)))

# =============================================================================
# >> Private Classes
# =============================================================================
            

class _PlayerCache(dict):
    def __missing__(self, key):
        self[key] = {
            'executions': 0,
            'execution': 0,
        }
        return self[key]
  

class _AdvancedCommand(object):
    _private_say_command_cache = None
    _console_command_cache = None
    _server_command_cache = None
    
    def __init__(self, kwargs):
        self._private_say_command_cache = {}
        self._console_command_cache = {}
        self._server_command_cache = {}
        for item in kwargs:
            setattr(self, item, kwargs[item])
    
    def _return(self, type):
        """
        Handles return the proper action (BLOCK, CONTINUE) if applicable for 
        this command instance
        """
        if self.args['type'] != 'public_say':
            return getattr(self, self.args['type'] + '_return_on_' + type)
        else:
            return
    
    def _get_log_args(self):
        """
        Formats some stuff for logging like the engine does
        """
        args = {
            'command': self.args['args'][0],
            'args': ', '.join(self.args['args'][1:])
        }
        
        if self.args['playerinfo'] is None:
            uargs = {
                'userid': 'Console',
                'name': 'Console',
                'steamid': 'Console',
            }
        else:
            p = self.args['playerinfo']
            uargs = {
                'userid': p.get_userid(),
                'name': p.get_name(),
                'steamid': p.get_networkid_string(),
            }
        
        args['user'] = '%(name)s<%(userid)s><%(steamid)s>' % uargs
          
        return args
    
    def _log_spam_protect(self):
        """
        Log method intended for logging when the spam protection kicks in
        """
        if (self.log_spam_protect and 
                self.log_logger.isEnabledFor(logging.INFO)):
            self.log_logger.info((
                '%(user)s was caught spamming command "%(command)s with ' 
                'arguments "%(args)s"') % self._get_log_args())
    
    def _log_spam_protect(self):
        """
        Log method intended for logging pre-command execution.
        """
        if self.log_pre_exec and self.log_logger.isEnabledFor(logging.INFO):
            self.log_logger.info((
                '%(user)s is trying to execute command "%(command)s with '
                'arguments "%(args)s"') % self._get_log_args())
    
    def _log_post_exec(self):
        """
        Log method intended for logging post-command execution.
        """
        if self.log_post_exec and self.log_logger.isEnabledFor(logging.INFO):
            self.log_logger.info((
                '%(user)s executed command "%(command)s" with arguments'
                '"%(args)s"') % self._get_log_args())
    
    def _check_spam_protection(self):
        """
        Perform all tasks necessary for spam protections.
        
        @return True if the command can be executed
        @return False if the user is eligible for spam protection
        """
        if not self.spam_protect:
            return True
        
        # Spam protection is turned off by default for server commands
        if not self.spam_protect_server:
            if self.args['playerinfo'] is None or self.args['type'] == 'server':
                return True
        
        userid = self.args['playerinfo'].get_userid()
        
        curtime = time()
        # Reset the timer and return if the difference is larger
        if (curtime - self.parent._players[userid]['execution'] >= 
                self.parent._spam_protect_delta):
            self.parent._players[userid]['execution'] = curtime
            self.parent._players[userid]['executions'] = 1
            return True
            
        self.parent._players[userid]['executions'] += 1
        if (self.parent._players[userid]['executions'] >= 
                self.spam_protect_attempts):
            # Actually log before actions incase the actions contain erros
            self._log_spam_protect()
            for action in self.spam_protect_action:
                action(self.args)
            return False
    
    def _check_access(self):
        """
        Checks whether the user has access to the command.
        
        @return True when the user does
        @return False otherwise
        """
        
        if not self.auth:
            return True
        
        # The server has access in any case
        if self.args['type'] == 'server':
            return True
        
        # Check for access
        if AuthManager.is_player_authorized(
                self.args['playerinfo'], *self.auth_params, 
                **self.auth_kwparams):
            return True
        else:
            for callback in self.auth_failure_callback:
                callback(self.args)
            return False
    
    def _execute(self, args):
        """
        Executes the command.
        
        @args - list of arguments passed to the callbacks
        """
        self.args = args
        self.args['command'] = self
        
        # First thing to do is to check the spam protection so we don't
        # waste extra cpu time if it was hit
        if not self._check_spam_protection():
            return self._return('failure')
        
        # Actually log that the command is going to be executed
        self._log_spam_protect()
        
        # Check auth unless the command was executed on the server
        if not self._check_access():
            return self._return('failure')
        
        # Check the pre_hook functions and stop execution if they fail
        for callback in self.pre_hook:
            if not callback(args):
                return self._return('failure')
        
        # Execute the callbacks
        for callback in self.callback:
            callback(args)
        
        # If the command was successfully executed, it will be logged
        self._log_post_exec()
        
        return self._return('success')


class _GenericCommandHandler():
    def __init__(self, type, command):
        self.type = type
        self.command = command
    
    def __call__(self, *args):
        """
        Handles command execution.
        
        Craft the command arguments such as userid, edict and type and change 
        the arguements into a list for easier use in python.
        """
        passargs = []
        
        for i in range(0, args[-1].get_arg_count()):
            passargs.append(args[-1].get_arg(i))
        
        cmdargs = {
            'args': passargs,
            'teamonly': None,
            'type': self.type,
        }
        
        if self.type == 'server':
            cmdargs['playerinfo'] = None
        elif self.type == 'console':
            cmdargs['playerinfo'] = args[0]
        elif self.type == 'private_say':
            cmdargs['playerinfo'] = args[0]
            cmdargs['teamonly'] = args[1]
        
        self.command._execute(cmdargs)

        
class _AdvancedCommandManager(object):
    """
    """
    def __init__(self):
        self._commands = {}
        self._public_say_command_cache = {}
        self._players = _PlayerCache()
        # 1 second works quite well I think
        self._spam_protect_delta = 1
    
    def _handle_public_say_command(self, event):
        """
        This method handles public say commands. 
        
        It is actually an event listener registered to player_say as public 
        commands are not handled though the engine directly.
        """
        text = event.get_string('text')
        
        # We will build the proper args later on, for now we just need the
        # command
        args = text.split(' ', 1)
        
        # Stop right here if this is not a registered command and save some
        # server performance
        if args[0] not in self._public_say_command_cache:
            return 
        
        cmd = self._public_say_command_cache[args[0]]['object']
        
        # Make sure there are arguments
        if len(args) == 2:
            # Parse this like the source engine does.
            # - Quoted strings are returned completly with the quotes 
            #   stripped
            # - Strings with a : between them are split up in the arg list
            match = re.findall(_saycmd_re, args[1:][0])
            args.pop(-1)
            for item in match:
                # Only one group matches at a time
                if item[0]:
                    # Remove the quotes
                    args.append(item[0].strip('"'))
                elif item[1]:
                    args.append(item[1])
                elif item[2]:
                    args.append(item[2])
        
        cmdargs = {
            'playerinfo': playerinfo_from_userid(event.get_int('userid')),
            'type': 'public_say',
            'args': args,
            'teamonly': event.get_int('teamonly'),
        }
        
        cmd._execute(cmdargs)
    
    def set_spam_protect_delta(self, delta=1):
        """
        Sets the time to use for spam protection.
        Your best bet is to leave it at the default.
        
        @delta - Time in seconds (Default: 1)
        """
        if delta <= 0:
            raise ValueError('Time delta must not be less then or equal to 0')
        self._spam_protect_delta = delta
    
    def register(
            self, 
            command, 
            callback=(), 
            description='No description given',
            # Spam Protection
            spam_protect = True, 
            spam_protect_attempts=5, 
            spam_protect_action=(_kick, ),
            spam_protect_server=False,
            # Command Aliases
            alias=(), 
            # Public say command settings
            public_say=False,
            public_say_prefix=(), 
            # Private say command settings
            private_say=False,
            private_say_prefix=(),            
            private_say_return_on_success=command_c.CommandReturn.CONTINUE,
            private_say_return_on_failure=command_c.CommandReturn.CONTINUE,
            # console command settings
            console=False, 
            console_prefix=(),
            console_return_on_success=command_c.CommandReturn.CONTINUE,
            console_return_on_failure=command_c.CommandReturn.CONTINUE,
            # Server Command settings
            server=False, 
            server_prefix=(),
            server_flags=cvar_c.FCVAR_NONE,
            server_return_on_success=command_c.CommandReturn.CONTINUE,
            server_return_on_failure=command_c.CommandReturn.CONTINUE,
            # Loging
            log_pre_exec=False,
            log_post_exec=False,
            log_spam_protect=False,
            log_logger=_logger,
            # Authentification
            auth = 0,
            auth_params = (),
            auth_kwparams = {},
            auth_failure_callback = (_access_denied, ),
            #
            help_enabled = False,
            # Pre Hook 
            
            pre_hook=()):
        """
        Registers the specified command.

        Notes:
        Any parameter that requires an iterable and does not receive one will 
        have the parameter automatically turned into a tuple. Be careful when 
        using strings, as they are iterable!
        
        In order to maintain compatibility with future cmdlib versions I 
        suggest you use keyword arguments instead of positional arguments.

        @command                       - Name of the command to register
        @callback                      - Iterable of callbacks for this command
        @description                   - Description for this command as 
                                         registered in the engine
        @spam_protect                  - Whether to enable spam protection
        @spam_protect_attempts         - Maximum attempts per second before 
                                         taking action
        @spam_protect_action           - Iterable of callables when the user 
                                         fails the spam protection checks
        @spam_protect_server           - Whether to enable spam protection for
                                         server
        @alias                         - Iterable of strings to use as command
                                         alias
        @public_say                    - Whether to register as public say 
                                         command
        @public_say_prefix             - Iterable of strings to use as prefix
                                         for public say commands
        @private_say                   - Whether to register as private say 
                                         command
        @private_say_prefix            - Iterable of strings to use as prefix 
                                         for private say commands
        @private_say_return_on_success - Internal return value of the command
                                         on successful execution
        @private_say_return_on_failure - Internal return value of the command
                                         on failed execution
        @console                       - Whether to register as console command
        @console_prefix                - Iterable of strings to use as prefix
                                         for console commands
        @console_return_on_success     - Internal return value of the command
                                         on successful execution
        @console_return_on_failure     - Internal return value of the command
                                         on failed execution
        @server                        - Whether to register as server command
        @server_prefix                 - Iterable of strings to use as prefix 
                                         for private commands
        @server_flags                  - Flags to use when creating the command
        @server_return_on_success      - Internal return value of the command 
                                         on successful execution
        @server_return_on_failure      - Internal return value of the command 
                                         on failed execution
        @log_pre_exec                  - Whether to log before the command 
                                         executes
        @log_post_exec                 - Whether to log after the command 
                                         executes
        @log_spam_protect              - Whether to log failed spam protection
                                         attempts
        @log_logger                    - Logger to use
        @auth                          - Auth type to use:
                                         0 = Disabled
                                          1 = Source Python authentication
        @auth_params                   - Positional parameters passed to the
                                         authentication function
        @auth_kwparams                 - Keyword parameter passed to the
                                         authentication function
        @auth_failure_callback         - Iterable of callables that will be 
                                         executed if authentication fails
        @help_enabled                  - TODO
        @pre_hook                      - Iterable of callables to execute 
                                         before the command runs. Only if all 
                                         return True the command is executed

        @raise KeyError if the command exists
        @raise TypeError if one of the callbacks supplied is not callable
        """
        # Can't register a command that exists :P
        if command in self._commands:
            raise KeyError('Command %s exists' % command)
        
        # Before doing anything, construct the argument dictionary :)
        args = {
            'parent': self,
        }
        l = locals()
        for item in l:
            if item in ('self', 'args', 'l', 'item'):
                continue
            args[item] = l[item]
        
        # Check whether these are iterable and if they are not make them :P
        for arg in (
                'callback', 'spam_protect_action', 'alias', 
                'public_say_prefix', 'private_say_prefix', 'console_prefix',
                'server_prefix', 'auth_failure_callback', 'pre_hook'):
            if not hasattr(args[arg], '__iter__'):
                args[arg] = (args[arg], )
            
        # Check for callables
        for arg in (
                'callback', 'spam_protect_action', 'auth_failure_callback', 
                'pre_hook'):
            for cb in args[arg]:
                if not hasattr(cb, '__call__'):
                    raise TypeError('Argument %s requires callables' % arg) 
         
        obj = _AdvancedCommand(args)
        
        # Build the cache
        cmds = list(alias) + [command, ]
        
        for type in ('public_say', 'private_say', 'console', 'server'):
            if not args[type]:
                continue
            prefixes = args[type + '_prefix']
            attrobj = self if type == 'public_say' else obj
            cache = getattr(attrobj, '_' + type + '_command_cache')
            if prefixes:
                for item in prefixes:
                    for cmd in cmds:
                        cache[item + cmd] = {'object': obj, 'callback': None}
            else:
                for cmd in cmds:
                    cache[cmd] = {'object': obj, 'callback': None}
        
        # Register private say commands:
        for cmdstr in obj._private_say_command_cache:
            c = command_c.get_say_command(cmdstr)
            cb = _GenericCommandHandler('private_say', obj)
            obj._private_say_command_cache[cmdstr]['callback'] = cb
            c.add_callback(cb)

        # Register console commands:
        for cmdstr in obj._console_command_cache:
            c = command_c.get_client_command(cmdstr)
            cb = _GenericCommandHandler('console', obj)
            obj._console_command_cache[cmdstr]['callback'] = cb
            c.add_callback(cb)
        
        # Register server commands
        for cmdstr in obj._server_command_cache:
            c = command_c.get_server_command(cmdstr, description, server_flags)
            cb = _GenericCommandHandler('server', obj)
            obj._server_command_cache[cmdstr]['callback'] = cb
            c.add_callback(cb)
        
        # Assign the object..
        self._commands[command] = obj
    
    def unregister(self, command):
        """
        Unregisters the specified command.
        
        @command - The name of the command to unregister
        
        @raise KeyError if the command does not exist
        """
        # Can't unregister a command that does not exist
        if command not in self._commands:
            raise KeyError('Command %s does not exist' % command)
        
        obj = self._commands[command]
        
        # Unregister private say commands:
        for cmdstr in obj._private_say_command_cache:
            c = command_c.get_say_command(cmdstr)
            c.remove_callback(
                obj._private_say_command_cache[cmdstr]['callback'])
        
        # Unregister console commands:
        for cmdstr in obj._console_command_cache:
            c = command_c.get_client_command(cmdstr)
            c.remove_callback(obj._console_command_cache[cmdstr]['callback'])
        
        # Unregister server commands:
        for cmdstr in obj._server_command_cache:
            c = command_c.get_server_command(cmdstr)
            c.remove_callback(obj._server_command_cache[cmdstr]['callback'])
        
        # Free the cache
        cmds = list(obj.alias) + [command, ]
        for type in ('public_say', 'private_say', 'console', 'server'):
            if not getattr(obj, type):
                continue
            prefixes = getattr(obj, type + '_prefix')
            attrobj = self if type == 'public_say' else obj
            cache = getattr(attrobj, '_' + type + '_command_cache')
            if prefixes:
                for item in prefixes:
                    for cmd in cmds:
                        del cache[item + cmd]
            else:
                for cmd in cmds:
                    del cache[cmd]
        
        # Everything is done, now remove the object
        del self._commands[command]
 
# =============================================================================
# >> INIT
# =============================================================================

AdvancedCommandManager = _AdvancedCommandManager()
_EventRegistry.register_for_event(
    'player_say', AdvancedCommandManager._handle_public_say_command)
