"""
Path     addons/source-python/packages/custom/k2tools/colormsg.py
Name     Color Tags
Version  1.00.000
Revision $Rev: 36 $
Author   (C) [#OMEGA] - K2


INFO

Adds ability to use tags to denote colours in chat without restrictions
imposed on the game type.


AGREEMENT
 
See k2tools_license.txt
"""

__k2doc__ = {'TypeManager': 'instance'}
__all__ = ['TypeManager', 'SayText2']
__version__ = '1.00.000 $Rev: 36 $'
__author__ = '[#OMEGA] - K2'
__date__ = '$Date: 2013-09-30 15:48:45 +0200 (Mo, 30 Sep 2013) $'

# =============================================================================
# >> Imports
# =============================================================================

# Python
import re

# Source.Python

from core import GAME_NAME
from messages import SayText2 as _SayText2

# K2Tools
from k2tools import config
from k2tools._paths import K2TOOLS_CFGSPEC_PATH
from k2tools._paths import K2TOOLS_DATA_PATH

# =============================================================================
# >> Globals
# =============================================================================

_cfg = config.load(
    K2TOOLS_DATA_PATH.joinpath('colormsg', GAME_NAME + '.ini'), 
    K2TOOLS_CFGSPEC_PATH.joinpath('colormsg-configspec.ini'))

_chex_re = re.compile(r'#chex[A-Fa-f0-9]{6}')
_cahex_re = re.compile(r'#cahex[A-Fa-f0-9]{8}')
_byte_re = re.compile(r'^[A-Fa-f0-9]{2}$')
    
# =============================================================================
# >> Classes
# =============================================================================


class _Type(dict):
    '''
    Special dictionary object with some extra attributes for proper type 
    parsing.
    '''
    def __init__(self):
        self.cahex = None
        self.chex = None
        self.default = None
        
        super(_Type, self).__init__()


class _TypeManager(dict):
    '''
    Prepares the data from the config files for use and provides functions for 
    parsing the tags
    '''
    def __init__(self):
        '''
        Adds SayText2 as message type by default
        '''
        super(_TypeManager, self).__init__()
        self._add_type('SayText2')
    
    def _convert(self, str):
        '''
        Convert the given byte string to the proper character
        
        @str - string to convert
        
        @return str appropriate character
        '''
        return bytes.fromhex(str).decode()
    
    def _add_type(self, type):
        '''
        Add a message Type for parsing.
        
        @type - Type to add
        '''
        self[type] = _Type()
        self[type].default = self._convert(_cfg[type]['default'])
        for item in _cfg[type]['Colors']:
            key = '#' + item
            if _cfg[type]['Colors'][item]:
                if re.match(_byte_re, _cfg[type]['Colors'][item]):
                    self[type][key] = self._convert(_cfg[type]['Colors'][item])
                else:  # Might be a color code instead
                    self[type][key] = _cfg[type]['Colors'][item]
            else:
                self[type][key] = self[type].default
        if _cfg[type]['rgb']:
            self[type].chex = self._convert(_cfg[type]['rgb'])
        if _cfg[type]['rgba']:
            self[type].cahex = self._convert(_cfg[type]['rgba'])
    
    def remove_tags(self, text, type='SayText2'):
        '''
        Removes all text for the given type for a string.
        
        @text - Text to remove the tags from
        @type - Type of the message that is used
        
        @return str formatted text
        '''
        for item in self[type]:
            text = text.replace(item, '')
        # Remove all them regardless of support
        for item in re.findall(_chex_re + '|' + _cahex_re, text):
            text = text.replace(item, '')
        return text
            
    def parse_tags(self, text, type='SayText2'):
        '''
        Parses the tags for the given type and returns the formatted text.
        
        @text - Text to parse the tags in
        @type - Type of message that is used
        
        @return str formatted text
        '''
        # Parse normal tags first. Note these might contain chex or cahex tags
        # as replacement
        for item in self[type]:
            text = text.replace(item, self[type][item])
        # RGB messages
        if _cfg[type]['rgb']:
            print(text)
            text = text.replace('#chex', self[type].chex)
        else:
            for item in re.findall(_chex_re, text):
                text = text.replace(item, '')
        
        # RGBA messages
        if _cfg[type]['rgba']:
            text = text.replace('#cahex', self[type].cahex)
        else:
            for item in re.findall(_cahex_re, text):
                text = text.replace(item, '')
        
        return text
   
# =============================================================================
# >> Functions
# =============================================================================


def SayText2(*args, **kwargs):
    '''
    Drop-in replacement for messages.SayText2
    
    Formats the message with the colors accordingly and returns a SayText2 
    object.
    
    @return SayText2 SayText2 instance
    '''
    if 'message' in kwargs:
        kwargs['message'] = TypeManager.parse_tags(
            kwargs['message'], 'SayText2')
    return _SayText2(*args, **kwargs)
   
# =============================================================================
# >> Init
# =============================================================================

TypeManager = _TypeManager()
