"""
Path     addons/source-python/packages/custom/k2tools/config.py
Name     General Purpose Config Tools
Version  1.00.000
Revision $Rev: 36 $
Author   (C) [#OMEGA] - K2


INFO

This library is just a collection of rather generic useful config classes and 
tools that I use in my scripts to make dealing with configs a little bit 
easier.


AGREEMENT

See k2tools_license.txt
"""

__all__ = ['ConfigError', 'load', 'update']
__version__ = '1.00.000 $Rev: 36 $'
__author__ = '[#OMEGA] - K2'
__date__ = '$Date: 2013-09-30 15:48:45 +0200 (Mo, 30 Sep 2013) $'


# =============================================================================
# >> Imports
# =============================================================================

# Python
from os.path import exists
from time import strftime

# Site Packages
from configobj import ConfigObj
from configobj import flatten_errors
from validate import Validator

# =============================================================================
# >> Private Functions
# =============================================================================


def _recursive(cfg, cfgspec):
    """
    TODO: Possibly hits recursion limit.
    """
    for item in cfgspec:
        if '__many__' in item:
            continue
        newcfgspec = cfgspec[item]
        newcfg = cfg[item]
        if isinstance(newcfgspec, dict) and len(list(newcfgspec.keys())):
            newcfg.comments = newcfgspec.comments
            newcfg.inline_comments = newcfgspec.inline_comments
            newcfg = _recursive(newcfg, newcfgspec)
        cfg[item] = newcfg
    return cfg

# =============================================================================
# >> Public Functions
# =============================================================================
 

def update(cfg, cfgspec=None):
    """
    Updates the specific config files while also keeping the comments
    
    @cfg     - ConfigObj instance
    @cfgspec - ConfigObj instance to act as config specification
    """
    if not cfgspec:
        cfgspec = cfg.configspec
    cfg.initial_comment = cfgspec.initial_comment
    cfg.final_comment = cfgspec.final_comment
    cfg = _recursive(cfg, cfgspec)
    cfg.write()


def load(cfgpath, validatepath, errorlogfile=''):
    """
    Loads the specified skill config and runs config verification.

    @cfgpath      - Path of the config file
    @validatepath - Path of the validation file
    @errorlogfile - Path of the error log file
    
    @return ConfigObj on success
    
    @raise IOError if the validation file can not be found
    @raise ConfigError on failure
    """
    if not exists(validatepath):
        raise IOError('Validation file does not exist')

    validator = Validator()

    config = ConfigObj(cfgpath, configspec=ConfigObj(
        validatepath, encoding='utf-8', raise_errors=True, file_error=True,
        _inspec=True), encoding='utf-8')
    errors = flatten_errors(
        config, config.validate(validator, preserve_errors=True, copy=True))
    
    if not errors:
        return config
    
    if errorlogfile:
        string = [
            'Config Verification (' + strftime('%d.%m.%Y %H:%M:%S') + ') \n', ]
        for error in errors:
            string.append('[' + config.filename + '] ')
            print(error)
            # Section
            if error[0]:
                for section in error[0]:
                    string.append('[' + section + '] ')
            else:
                string.append('[missing section] ')
            
            # Key
            if error[1]:
                string.append('[' + error[1] + '] ')
            else:
                string.append('[missing key] ')
                
            # Value
            if error[2]:
                string.append(str(error[2]))
            else:
                string.append('missing value)')
            string.append('\n')
        string.append('\n')
        with open(errorlogfile, 'a') as f:
            f.write(''.join(string))
        raise ConfigError(
            'Config parsing failed. See %s for details.' % errorlogfile)
    else:
        raise ConfigError('Config parsing failed.')
            
# =============================================================================
# >> Classes
# =============================================================================


class ConfigError(Exception):
    """
    Exception used when a config error is encountered
    """
    pass
