"""
Path     addons/source-python/packages/custom/k2tools/keyvalues.py
Name     Keyvalues
Version  1.00.000
Revision $Rev: 37 $
Author   (C) [#OMEGA] - K2


IMPORTANT

Intended for Valve KeyValue format.

See this page for details:
https://developer.valvesoftware.com/wiki/KeyValues_class


INFO

Parser for Valve's KeyValues format. It also supports some non-valve features
like multi-line comments.


AGREEMENT
 
See k2tools_license.txt

Otherwise should be noted that the KeyValues Format is (C) Valve Corporation


IDEAS/TODO:

- Support #include/#base for KeyValuesFile
- Support reading/writing comments

"""

__all__ = ['ParseError', 'KeyValues', 'KeyValuesParser', 'KeyValuesFile']
__version__ = '1.00.000 $Rev: 37 $'
__author__ = '[#OMEGA] - K2'
__date__ = '$Date: 2013-10-01 07:29:50 +0200 (Di, 01 Okt 2013) $'

# =============================================================================
# >> Imports
# =============================================================================

import re
from os.path import exists

# =============================================================================
# >> Globals
# =============================================================================
_keygroup_re = re.compile(r'("(?:[^"\t]*)")|([^"\s{}]+)|({)|(})')

# =============================================================================
# >> Classes
# =============================================================================


class ParseError(Exception):
    pass


class KeyValues(object):
    """
    The most basic KeyValues class which acts as KeyValue representation in
     python.
    
    It works similar to a dictionary in python, however, all keys are ordered.
    """
    def __init__(self, parent=None):
        """
        Crates a new KeyValues object
        
        @parent - parent KeyValues object (if present)
        
        @raise TypeError if parent is not a KeyValues object
        """
        self._data = {}
        self._odata = []
        if parent and not isinstance(parent, KeyValues):
            raise TypeError("Parent must be KeyValues object")
        self.parent = parent
    # Container type emulation
    
    def __getitem__(self, key):
        """
        Returns the value of the specified key.
        I.e.: kv[key] -> value
        
        @return object -
        """
        return self._data[key]
    
    def __setitem__(self, key, value):
        """
        Sets the value of the specified key.
        I.e.: kv[key] = value
        """
        if key not in self._data:
            self._odata.append(key)
        self._data[key] = value
    
    def __delitem__(self, key):
        """
        Removes the specified key
        I.e.: del kv[key]
        """
        del self._data[key]
        self._odata.pop(self._odata.index(key))
    
    def __contains__(self, key):
        """
        Returns whether key is contained in the object.
        I.e.: key in kv
        
        @return True if found
        @return False if not
        """
        return key in self._odata
    
    # Iterator
    def __iter__(self):
        """
        Returns itself as Iterator and sets the beginning of the iteration 
        properly.
        
        @return KeyValues returns itself as Iterator
        """
        self._item = 0
        return self
    
    def __next__(self):
        """
        @return object next key in the iteration
        """
        if self._item >= len(self._odata):
            raise StopIteration()
        self._item += 1
        return self._odata[self._item - 1]
    
    # Representation
    def __repr__(self):
        """
        Same functionality as __str__ currently.
        I.e.: repr(kv)
        
        @return str -
        """
        return str(self)
    
    def __str__(self):
        """
        Returns this KeyValues object as string representation.
        For subkeys, this is done recursively.
        I.e.: str(kv)
        
        @return str -
        """
        vstr = []
        for key in self._odata:
            vstr.append('"' + str(key) + '" ')
            
            value = self._data[key]
            if isinstance(value, KeyValues):
                vstr.append('{ ' + str(value) + ' } ')
            else:
                vstr.append('"' + str(value) + '" ')
        return ''.join(vstr)
    
    # Basic Methods
    def __eq__(self, obj):
        """
        Checks whether this object is equal to another.
        I.e.: kv == obj
        
        @return NotImplemented if object is not a KeyValues object
        @return True if object has matching data 
        @return False otherwise
        """
        if not isinstance(obj, KeyValues):
            return NotImplemented
        if self._odata == obj._odata and self._data == obj._data:
            return True
        return False
    
    def __ne__(self, obj):
        """
        Reverse of __eq__.
        I.e: kv != obj
        
        @return True if objects are not KeyValues objects
        @return True don't contain the same data
        @return False otherwise
        """
        return not self.__eq__(obj)
    #
    # Public methods
    #

    # Compability with dicts
    def deep_copy(self):
        """
        Provides a deep independent copy of this KeyValue object.
        
        The difference to .copy() is that values that are KeyValue objects
        will be recreated instead of staying the same object.
        
        However, objects of other types, will still have the same problem,
        though "officially" you should not assign any non basic types as value.
        
        @return KeyValues deep-copy of this object
        """
        tmp = self.copy()
        for key in tmp:
            value = tmp[key] 
            if isinstance(value, KeyValues):
                tmp[key] = value.deep_copy()
        return tmp
    
    def copy(self):
        """
        Provides a superficial independent copy of this KeyValue object.
        
        Keys that do have a KeyValues object as value will still point to the 
        same object in memory (which means, changing such an KeyValue will
        affect both the copy and the original object).
        
        @return KeyValues superficial copy of this object
        """
        # This should even work for subclasses
        tmp = self.__class__()
        tmp._data = dict(self._data)
        tmp._odata = list(self._odata)
        return tmp
    
    def values(self):
        """
        Returns an ordered list of values.
        
        @return list an ordered list of values.
        """
        tmp = []
        for item in self._odata:
            tmp.append(self._data[item])
        return tmp
    
    def keys(self):
        """
        Returns an ordered list of keys.
        
        @return list an ordered list of keys.
        """
        return list(self._odata)
    
    def merge(self, obj, override=True):
        """
        Merges this KeyValues object with another KeyValues object.
        
        @obj      - KeyValues object to merge with
        @override - Whether to override Key/Value pairs if they exist in the
                    other object.
                    True will use the new value from the merging object
                    False will use the original value from this object
        """
        if not isinstance(obj, KeyValues):
            raise TypeError("Obj must be a KeyValues object")
        for key in obj.keys():
            value = obj[key]
            if key in self:
                if isinstance(value, KeyValues):
                    if isinstance(self[key], KeyValues):
                        self[key].merge(value, override)
                    elif override:
                        value.deep_copy()
                elif override:
                    self[key] = value
            else:
                if isinstance(value, KeyValues):
                    self[key] = value.deep_copy()
                else:
                    self[key] = value
    
    def get_depth(self):
        """
        Returns the current depth of this KeyValues object in a Key/Subkey 
        hirarchy.
        
        0 is the top-level.
        
        @return int depth
        """
        level = 0
        obj = self
        while (obj.parent):
            level += 1
            obj = obj.parent
        return level
    
    def walk(self, depth=-1):
        """
        Walks though this object and returns key-value pairs
        
        @depth - If known, the depth can be specified here (Default: -1)
                 If set to -1, the depth will be automatically determined
        
        @return list a list containing the following tuple:
        (KeyValues object, key, value, depth)
        """
        # Determine level
        if depth == -1:
            depth = self.get_depth()
            
        tmp = []
        for key in self._odata:
            value = self[key]
            tmp.append((self, key, value, depth))
            if isinstance(value, KeyValues):
                tmp += value.walk(depth + 1)
        return tmp
            
       
class KeyValuesParser(KeyValues):
    """
    This class adds parsing capabilities on top of the base KeyValues class.
    """
    def parse(self, data="", typecast=False):
        """
        Parses the given data into the current object - resets all previous
        data.
        
        @data    - A string containing data to parse with the KeyValues syntax 
        """
        # Init variables
        self._data = {}
        self._odata = []
        self._typecast = typecast
        self._pointer = self
        self._lastkey = None
        self._iskey = True
        
        # Start parsing
        
        # Non-standard: Remove multi-line/inline comments
        comments = [0]
        running = True
        char = 0
        while (running):
            char = data.find('/*', char)
            if char != -1:
                comments.append(char)
                end = data.find('*/', char)
                if end == -1:
                    raise ParseError(
                        "No closing comment for multiline at pos %s" % char)
                # real end is after */ and not where it starts
                comments.append(end + 2)
                char = end
            else:
                comments.append(len(data))
                running = False

        newdata = ''
        for i in range(0, len(comments)):
            if i % 2 == 0:
                newdata += data[comments[i]:comments[i + 1]]
        
        newdata = newdata.replace('\r\n', '\n')
        newdata = newdata.replace('\r', '\n')
        for line in newdata.split('\n'):
            # Ingore comments
            if '//' in line:
                line = line[:line.index('//')]
            # Ingore empty lines
            if not line:
                continue
            # Parse the string
            matches = re.findall(_keygroup_re, line)
            for match in matches:
                # Group 0: Quoted key or value
                if match[0]:
                    self._add_item(match[0].strip('"'))
                # Group 1: key or value
                elif match[1]:
                    self._add_item(match[1])
                # Group 2: Start of subkey
                elif match[2]:
                    self._start_kv()
                # Group 3: End of subkey
                elif match[3]:
                    self._end_kv()
    
    def _start_kv(self):
        """
        Starts a new KeyValue pair
        """
        #if self._iskey:
        #    raise ParseError("Subkey can't act as a key")
        self._pointer[self._lastkey] = KeyValues(self._pointer)
        self._pointer = self._pointer[self._lastkey]
        # Next item must be a key again and can't be a value
        self._iskey = True
    
    def _end_kv(self):
        """
        Ends a new KeyValue pair
        """
        self._pointer = self._pointer.parent
    
    def _add_item(self, item):
        """
        Adds an item to the current KeyValue pair
        """
        if self._iskey:
            self._pointer[item] = None
            self._lastkey = item
        else:
            if self._typecast:
                for cast in (int, float):
                    try:
                        item = cast(item)
                        break
                    except:
                        continue
                
            self._pointer[self._lastkey] = item
            
        self._iskey = not self._iskey

    
class KeyValuesFile(KeyValuesParser):
    """
    The KeyValuesFile class handles using KeyValues with file input/output.
    """
    def __init__(self, *args, **kwargs):
        super(KeyValuesFile, self).__init__(*args, **kwargs)
        self.file = ''
    #
    # Private methods
    #
    
    def _recursive(self, obj, level=0):
        outstr = []
        tab = '\t' * level
        for key in obj._odata:
            outstr.append(tab + '"' + str(key) + '"')
            
            value = obj._data[key]
            if isinstance(value, KeyValues):
                outstr.append('\n' + tab + '{\n')
                outstr.append(self._recursive(value, level + 1))
                outstr.append(tab + '}\n')
            else:
                outstr.append(' "' + str(value) + '"\n')
        return ''.join(outstr)
   
    #
    # Public methods
    #
    def load(self, file=None, typecast=False):
        """
        Loads the specified file and parses the KeyValues data contained
        within.
        
        @file    - Path to the KeyValue file
        @newline - If None let python determine the newline type
        """
        if not file:
            file = self.file
            
        with open(file, "r") as f:
            self.parse(f.read(), typecast)
            # Stuff worked - set this as file
            self.file = file
    
    def write(self, file=None):
        """
        Writes the current data stored in this object to the specified file in
        KeyValues format. Please note that this will override any existing
        file.
        
        @file - Path to the KeyValue file
        """
        if not file:
            file = self.file
        
        with open(file, "w") as f:
            outstr = ['// Generated by keyvalues.py\n\n', ]
            outstr.append(self._recursive(self))
            f.write(''.join(outstr))
