"""
Path     addons/source-python/packages/custom/k2tools/popup/manager.py
Name     Popup Manager
Version  1.00.000
Revision $Rev: 39 $
Author   (C) [#OMEGA] - K2


INFO

Holds objects related to managing active popups and re-queueing them to
display.


AGREEMENT
 
See k2tools_license.txt
"""

# =============================================================================
# >> Imports
# =============================================================================

# Source Python C++
import command_c
from conversions_c import index_from_playerinfo
from player_c import PlayerGenerator

# Source Python
from events.manager import EventRegistry 
from messages import ShowMenu

# Python
from threading import Thread
from time import sleep

# K2Tools
from k2tools.threadlib import game_thread

# Popup
from .constants import *

# =============================================================================
# >> Script Information
# =============================================================================

__all__ = ['PopupManager']
__k2doc__ = {'PopupManager': 'instance'}
__version__ = '1.00.000 $Rev: 39 $'
__author__ = '[#OMEGA] - K2'
__date__ = '$Date: 2014-02-09 22:15:20 +0100 (So, 09 Feb 2014) $'

# =============================================================================
# >> Script Information
# =============================================================================


# Set on package init
class Popup:
    pass


class PopupList: 
    pass


class PopupMenu: 
    pass

# =============================================================================
# >> Classes
# =============================================================================


class _PopupRefresher(Thread):
    def __init__(self, *args, **kwargs):
        self._running = True
        super(_PopupRefresher, self).__init__(*args, **kwargs)
    
    def run(self):
        while self._running:
            game_thread(PopupManager._refresh_popups)
            sleep(TIMEOUT)


class _UserEntry(dict):
    def __init__(self):
        """
        Set defaults
        """
        # Display Data
        self['opts'] = []
        self['title'] = ''
        self['description'] = ''
        #
        self['init_time'] = 0
        self['index'] = 0
        self['language'] = None
        self._set_inactive()
    
    def _set_inactive(self):
        """
        Set to no active popups for this user entry.
        """
        self['active'] = False
        self['duration'] = 0
        self['pageid'] = 0
        self['popup'] = None

            
class _PopupManager(dict):
    """
    Holds a cache for each user, as well as event listeners and provides some
    utility functions.
    """
    def __init__(self):
        """
        Creates the PopupManager instance.
       
        Adds all current players to the cache.
        """
        for playerinfo in PlayerGenerator():
            self._add_player(playerinfo.get_userid())
    
    def _callback_handler(self, playerinfo, command):
        """
        client command filter
        
        Handles "menuselect" universially for each popup so they work
        correctly.
        """
        # Otherwise we'll be capturing all commands
        if command.get_arg(0) != 'menuselect':
            return    
            
        # Bots should not execute commands
        if playerinfo.is_fake_client():
            return
        
        userid = playerinfo.get_userid()
        
        # No popup is running at the moment - ignore
        if not self[userid]['active']:
            return
        
        option = int(command.get_arg(1))
        
        # Option handling
        if option == OPTION_PREVIOUS and \
                isinstance(self[userid]['popup'], PopupList):
            # Possible out of bounds will be fixed upon message formatting
            self[userid]['pageid'] -= 1
            self[userid]['popup']._display(self[userid])
        elif option == OPTION_NEXT and \
                isinstance(self[userid]['popup'], PopupList):
            # Possible out of bounds will be fixed upon message formatting
            self[userid]['pageid'] += 1
            self[userid]['popup']._display(self[userid])
        elif option == OPTION_CLOSE:
            self[userid]._set_inactive()
        elif OPTION_MIN <= option <= OPTION_MAX and \
                isinstance(self[userid]['popup'], PopupMenu):
            # Lists start at 0, not 1:
            option -= 1
            
            # Get the option
            offset = self[userid]['pageid'] * \
                self[userid]['popup']._options_per_page + option
            # Return if the option is invalid
            try:
                optdict = self[userid]['opts'][offset]
            except IndexError:
                return
                
            opt = optdict['option']
            
            if not opt._callback:
                return
            
            # The option is not active
            if not optdict['active']:
                return
            
            if callable(opt._callback):
                popup = self[userid]['popup']
                # Hide the current popup
                self.unsend(playerinfo)
                opt._callback(
                    playerinfo, opt, popup, *opt._callback_args, 
                    **opt._callback_kwargs)
            elif isinstance(opt._callback, Popup):
                opt._callback.send(
                    playerinfo, *opt._callback_args, **opt._callback_kwargs)
        
    def _add_player(self, userid, **kwargs):
        """
        Adds a new player to the popup cache.
        
        @userid - userid to create a new entry for
        @kwargs - arguments to add or update to the user dictionary
        """
        self[userid] = _UserEntry()
        self[userid].update(**kwargs)
    
    def _player_connect(self, event):
        """
        player_connect event hook
        """
        userid = event.get_int('userid')
        self._add_player(userid)
    
    def _player_disconnect(self, event):
        """
        player_disconnect event hook
        """
        userid = event.get_int('userid')
        del self[userid]
    
    def _refresh_popups(self):
        """
        Refreshes the popups from the cache.
        """
        for userid in self:
            if self[userid]['active']:
                self[userid]['popup']._display(self[userid])
    
    def get_active_popup(self, playerinfo):
        """
        Returns the currently active popup for the specified user
        
        @playerinfo - the CPlayerInfo instance to retrieve the popup for
        
        @return None if no popup is currently active
        @return Popup otherwise
        """
        p = self[playerinfo.get_userid()]
        if not p['active']:
            return None
        
        return p['popup']
    
    def unsend(self, playerinfo):
        """
        Hide whatever popup is currently active for the specified user
        
        @playerinfo - the CPlayerInfo instance of the user to hide the popup 
                      for
        """
        self[playerinfo.get_userid()]._set_inactive()
        
        ShowMenu(
            slots=0b1111111111,
            time=0, 
            message='',
        ).send(index_from_playerinfo(playerinfo))
            
# =============================================================================
# >> Init
# =============================================================================
          
PopupManager = _PopupManager()

PopupRefresher = _PopupRefresher()
PopupRefresher.start()
            
command_c.register_client_command_filter(PopupManager._callback_handler)
EventRegistry.register_for_event(
    'player_connect', PopupManager._player_connect)
EventRegistry.register_for_event(
    'player_disconnect', PopupManager._player_disconnect)
