"""
Path     addons/source-python/packages/custom/k2tools/popup/options.py
Name     Popup
Version  1.00.000
Revision $Rev: 36 $
Author   (C) [#OMEGA] - K2


INFO

Holds classes to serve as option for popups.


AGREEMENT
 
See k2tools_license.txt
"""

# =============================================================================
# >> Script Information
# =============================================================================

__all__ = ['PopupOption']
__version__ = '1.00.000 $Rev: 36 $'
__author__ = '[#OMEGA] - K2'
__date__ = '$Date: 2013-09-30 15:48:45 +0200 (Mo, 30 Sep 2013) $'

# =============================================================================
# >> Script Information
# =============================================================================


# Set on package init
class Popup: 
    pass

# =============================================================================
# >> CLASSES
# =============================================================================


class PopupOption(object):
    """
    Representation of an Option within a popup.
    """
    _attr = (
        '_name', '_value', '_translate', '_active', '_callback', 
        '_callback_args', '_callback_kwargs', '_translate_callback', 
        '_prehook')
    
    def __init__(
            self, name, value='', translate=True, active=None, 
            callback=None, callback_args=(), callback_kwargs={}, 
            translate_callback=None, prehook=None):
        """
        Creates a new popup option instance which is used for option handling
        in popups. 
        
        NAME:
        
        The name of an option must unique within a singe popup, you can use an
        option in multiple popups, however. If value is not specified, the name
        is used as text in the popup output formatting, and if translate is set
        to true, it will serve as the translation id.
        
        
        VALUE:
        
        You can set value if you want a different str value that is not
        translated - this use useful for displaying player names and similar
        things.
        
        
        TRANSLATE:
        
        If set to True, this option will be translated with the name of the 
        option as Translation ID. Please verify that the ID exists in
        the translation file you've specified for your Popup instance,
        otherwise you'll get errors when building/displaying the Popup.
        
        
        TRANSLATE_CALLBACK:
        
        This option can be additionally be specified to allow retrieving
        arguments on the fly when parsing the option.
        
        The callback specified must take 3 arguments:
        
        - 1. playerinfo - The CPlayerInfo instance of the user
        - 2. option     - This option instance
        - 3. popup      - The Popup instance
        - 4. language   - The language of the user
        
        ACTIVE:
        
        If set to True, this Option will be marked as active; this means
        it will appear yellow in game. However, this only works if the
        option starts with a number terminated by a dot (i.e. "123.").
        Note that permissions from prehook may override this behaviour.
        Also, if not set explicitly and if a callback is specified, True
        will be assumed.
        
        
        CALLBACK:
        
        Callback can be either a method to execute when this option is chosen
        or a Popup instance to display. Please note that callbacks are only
        handled when using PopupMenu.
        
        The callback specified must take 3 arguments;
        - 1. playerinfo - The CPlayerInfo instance of the user
        - 2. option     - This option instance
        - 3. popup      - The Popup instance
       
        
        You can specify additional arguments to pass with callback_args and
        callback_kwargs; it is suggested you use these if you specified
        a popup instance as callback - they'll be passed to the popup's
        send method.
        
        
        PREHOOK:
        
        The prehook must a valid callable that takes 3 arguments:
        - 1. playerinfo - The CPlayerInfo instance of the user 
        - 2. option     - This option instance
        - 3. popup      - The Popup instance
        - 4. language   - The language of the user
        
        This method is intended for access control within a popup. You can
        return one of these constants:
        
        ACCESS_ALLOWED 
        * Default Behaviour
        
        ACCESS_RESTRICTED
        * Shows the option as inactive, regardless of active settings
        
        ACCESS_DENITED
        * Completely hides the option for the current user from the popup
        
        
        @name               - Unique name of the PopupOption
        @value              - String value to display; if empty, name is used
        @translate          - Whether this option is translate able
        @translate_callback - Called when trying to translate
        @active             - Whether this option is active (-> added)
        @callback           - Callback or Popup instance
        @callback_args      - Arguments to pass to the callback/popup
        @callback_kwargs    - Keyword arguments to pass to the callback/popup
        @prehook            - Method 
        
        @raise TypeError when callback is not callable or not a popup instance
        @raise TypeError when prehook is not callable
        """
        self._name = name
        self._value = value
        self._translate = translate
        self._callback_args = callback_args
        self._callback_kwargs = callback_kwargs 
        self._set_callback(callback)
        self._set_callable('translate_callback', translate_callback)
        self._set_callable('prehook', prehook)
        
        if active is None:
            self._active = bool(self._callback)
        else:
            self._active = active
    
    def __str__(self):
        return 'PopupOption(%s)' % ', '.join([
            "%s=%s" % (at, getattr(self, at)) for at in self._attr])
    
    def _set_callback(self, callback):
        if callback is not None and not callable(callback) and \
                not isinstance(callback, Popup):
            raise TypeError(
                'callback must be a callable or another popup instance')
            
        self._callback = callback
    
    def _set_callable(self, name, value):
        if value is not None and not callable(value):
            raise TypeError('%s must be a callable' % name)
        
        setattr(self, '_' + name, value)
