"""
Path     addons/source-python/packages/custom/k2tools/popup/popup.py
Name     Popup Classes
Version  1.00.000
Revision $Rev: 39 $
Author   (C) [#OMEGA] - K2


INFO

This file holds the various available popup classes. 

Popup

Base class for all other popups. It is a very basic class that only displays
a single page with (theoretically) infinite amount of options. Suitable if you
only want to display a little bit of information.
This should be preferred over messages.ShowMenu to avoid messing things up.

Also see popups.Popup for more details


PopupList

This class supports displaying pages; if you want to display anything with a
list-like character, such as ranking lists this is a good option to start.
Unlike PopupMenu, the amount of items per page is (theoretically) unrestricted.

Also see popups.PopupList for more details


PopupMenu

This class is intended for a typical menu usage and is the only one that 
supports callbacks on the handler.
You should choose this class if you want the popup to interact with your
script instead of just displaying information

Also see popups.PopupMenu for more details



AGREEMENT
 
See k2tools_license.txt
"""

# =============================================================================
# >> Imports
# =============================================================================

# Source Python C++
from conversions_c import index_from_playerinfo
from player_c import PlayerGenerator

# Source Python PY
from core import GameEngine
from messages import ShowMenu
from translations.manager import LanguageManager
from translations.strings import LangStrings

# Python
import logging
from collections import OrderedDict
from time import time

# K2Tools

from k2tools._shared import lang

# Popup
from .constants import *
from .manager import PopupManager
from .options import PopupOption

# =============================================================================
# >> Script Information
# =============================================================================

__all__ = ['Popup', 'PopupList', 'PopupMenu']
__version__ = '1.00.000 $Rev: 39 $'
__author__ = '[#OMEGA] - K2'
__date__ = '$Date: 2014-02-09 22:15:20 +0100 (So, 09 Feb 2014) $'

# =============================================================================
# >> Globals
# =============================================================================

_log = logging.getLogger('k2tools.popup.popups')

# =============================================================================
# >> Classes
# =============================================================================

   
class Popup(object):
    """
    Basic Popup object. All popups should inherit from this base class.
    
    
    ADDING LINES: 
    
    It behaves similar to a list/dict hybrid, with the exception that it only 
    accepts PopupOptions (or strings, which then will be converted to a 
    PopupOption) as entries.
    Also see options.PopupOption for details.
    
    Also the functionality is more limited then a list, see the available
    methods for details, but it still acts as Iterator and container-type
    object.
    
    
    ATTRIBUTES / CACHE
    
    When using popups you should not be changing any internal attributes.
    In addition, after a popup has been send, you should not be editing the
    popup, as it will lead to unexpected behaviour; the popups currently 
    implement some sort of cache which is created upon sending, so that
    when a popup has to be re-displayed for a specific user, it doesn't have
    to go though the expensive process of translating, using callbacks, 
    re-sorting and so on.
    
    
    SUB-CLASSING
    
    If you want to write your own Popup class you should be sub-classing this
    class and override the formatting methods. Note that when messing with
    options as pages it is required to sub-class PopupList or PopupMenu so the
    handler catches pages properly.
    """
    def __init__(self, title=None, description=None, translations=None):
        """
        Creates a new Popup instance.
        
        @title        - Title of the popup; see Popup.set_title
        @description  - Description of the popup; see Popup.set_description
        @translations - LangStrings instance to use when translating options
        
        @raise TypeError if the title is neither PopupOption, str nor None
        """
        if translations and not isinstance(translations, LangStrings):
            raise TypeError('translations must be a valid LangStrings object')
        self._translations = translations
        self._options = []
        self._options_str = {}
        self.set_title(title)
        self.set_description(description)
        self.set_separator(SEPARATOR_BODY, '-----------------')
        self.set_separator(SEPARATOR_TITLE, ' ')
    #
    # Container Type Emulation
    #
    
    def __contains__(self, option):
        """
        Allows for checking directly whether an option in the popup.
        
        Examples
        'unique' in mypopup
        PopupOption('unique') in mypopup
        
        @return True if the option is in the popup
        @return False if the option is not in the popup
        @return False if the option is neither str nor PopupOption
        """
        if isinstance(option, PopupOption):
            return option._name in self._options_str
        elif isinstance(option, str):
            return option in self._options_str
        else:
            return False
    
    def __delitem__(self, key):
        """
        Allows for deleting options directly with a key or an index value.
        
        Examples:
        del mypopup['myoption']
        del mypopup[0]
        
        @key - Str or Int key
        
        @raise ValueError if a str key isn't in the options
        @raise IndexError if an int key is out of bounds
        """
        if isinstance(key, str):
            self.remove(key)
        elif isinstance(key, int):
            self.pop(key)
    
    def __getitem__(self, key):
        """
        Allows for getting options directly with a key or an index value.
        
        Examples:
        mypopup['myoption']
        mypopup[0]
        
        @key - Str or Int key
        
        @return PopupOption the option
        
        @raise IndexError if an int key isn't in the options
        @raise KeyError if a str key isn't in the options
        """
        if isinstance(key, str):
            return self._options_str[key]
        elif isinstance(key, int):
            return self._options[key]
    
    def __setitem__(self, key, option):
        """
        Allows for setting options directly with a key or an index value.
        
        Examples:
        mypopup['new'] = PopupOption('new')
        mypopup['new'] = 'new'
        mypopup[0] = PopupOption('override')
        
        @key    - Str or Int key
        @option - Option to set, either a str or PopupOption
        
        @raise IndexError if an int key is out of bounds
        @raise KeyError if a str key doesn't match the option name
        @raise KeyError if the option already exists
        @raise KeyError if the option is translatable but not in translations
        @raise TypeError if the option is neither PopupOption nor str
        """
        # Make this a valid option and raise errors if necessary
        option = self._create_option(option)
        
        if isinstance(key, str):
            # Make sure it doesn't mess up the uniqueness
            if key != option._name:
                raise KeyError(
                    'The key and name of the option must be the same')
            # Does this override an existing option?
            if key in self._options_str:
                # If so remove the old option first
                self._options.remove(self._options_str[key])
            self._options.append(option)
            self._options_str[key] = option
        elif isinstance(key, int):
            # Get old option at a given key, raises IndexError if out of bounds
            old = self._options[key]
            # Is this really Unique? 
            if option._name in self._options_str:
                # We can replace the option at the given index, but only if the
                # name is the same
                if self._options_str[option._name]._name != option._name:
                    raise('The option exists already')
            del self._options_str[old._name]
            self._options[key] = option
            self._options_str[option._name] = option
         
    def __bool__(self):
        """
        Always true.
        
        Override as it would be using __len__ otherwise (which ma return false
        on popups with empty options but an title or description).
        
        @return bool 
        """
        return True
    
    def __len__(self):
        """
        Returns the amount of popup options.
        
        Example:
        len(mypopup)
        
        @return int amount of options
        """
        return len(self._options)
    
    def __iter__(self):
        """
        Returns an iterator over this popup.
        
        @return Popup this popup instance as iterator
        """
        self._iter_index = -1
        self._iter_maxindex = len(self)
        return self
    
    def __next__(self):
        """
        Returns the next element in the popup.
        
        @return PopupOption current element
        
        @raise StopIterator when iteration maxlength is reached
        """
        self._iter_index += 1
        if self._iter_index >= self._iter_maxindex:
            raise StopIteration
        return self._options[self._iter_index]
    #
    # getters/setters
    #
    
    def set_description(self, description=None):
        """
        Sets the description of the popup.
        
        You can use None to clear the description.
        
        @description - the description to set
        
        @raise TypeError if the description is neither PopupOption, str nor
                         None
        """
        if description is None:
            self._description = description
        else:
            self._description = self._create_option(description)
    
    def set_title(self, title=None):
        """
        Sets the title of the popup.
        
        You can use None to clear the title.
        
        @title - the title to set
        
        @raise TypeError if the title is neither PopupOption, str nor None
        """
        if title is None:
            self._title = title
        else:
            self._title = self._create_option(title)
    
    def set_separator(self, location=SEPARATOR_ALL, separator=None):
        """
        Sets the separator of the popup.
        
        LOCATION:
        
        One of the SEPARATOR constants defined in popup.constants
        
        
        @location  - which separator to set
        @separator - separator to set, None to set empty
        """
        if not separator:
            separator = ''
        if location & SEPARATOR_TITLE:
            self._separator_title = separator
        if location & SEPARATOR_TOP:
            self._separator_top = separator
        if location & SEPARATOR_BOTTOM:
            self._separator_bottom = separator
        if location & SEPARATOR_CLOSE:
            self._separator_close = separator
    
    #
    # Private
    #
    
    def _get_id(self, value):
        """
        Returns the id (unique name) of the given string or PopupOption.

        @value - the value to retrive the id for
        
        @return str if it's a valid str or PopupOption
        @return None otherwise
        """
        if isinstance(value, PopupOption):
            return value._name
        elif isinstance(value, str):
            return value
        else:
            return None
    
    def _create_option(self, option):
        """
        Creates a valid PopupOption from a given string or PopupOption and
        raises error on failure. 
        
        @option - option to create from this string or PopupOption
        
        @return PopupOption if successful
        
        @raise KeyError if the option is translatable but not in translations
        @raise TypeError if the option is neither PopupOption nor str
        """
        if isinstance(option, str):
            option = PopupOption(option)
        elif not isinstance(option, PopupOption):
            raise TypeError('Option must be PopupOption or string')
            
        if self._translations and option._translate:
            id = option._value if option._value else option._name
            if not id in self._translations:
                raise KeyError(
                    'Option with name=%s, value=%s, translate=True is not in'
                    ' the LangStrings object' % (option._name, option._value))
            
        return option
    
    def _create_valid_option(self, option):
        """
        Creates a valid PopupOption from a given string or PopupOption and
        raises error on failure. 
        
        @option - option to create from this string or PopupOption
        
        @return PopupOption if successful
        
        @raise KeyError if the option already exists
        @raise KeyError if the option is translatable but not in translations
        @raise TypeError if the option is neither PopupOption nor str
        """
        option = self._create_option(option)
        
        if option._name in self._options_str:
            raise KeyError('Option %s already exists' % option)
        
        return option
    #
    # List/Dict Emulation
    #
    
    def append(self, option):
        """
        Appends a new option to the end of options.
        Each option must have an unique name.
        
        @option - option to append 
        
        @return PopupOption if appending was successful
        
        @raise KeyError if the option already exists
        @raise KeyError if the option is translatable but not in translations
        @raise TypeError if the option is neither PopupOption nor str
        """
        option = self._create_valid_option(option)
        
        self._options.append(option)
        self._options_str[option._name] = option
    
    def clear(self):
        """
        Removes all entries from the popup.
        """
        self._options.clear()
        self._options_str.clear() 
    
    def index(self, option):
        """
        Returns the index of the given option.
        
        @option - option to retrieve the index for
        
        @return int index of the option if present
        
        @raise ValueError if the options doest not exist
        """
        # list raises ValueError, so it's for consistency - instead of KeyError
        if option not in self:
            raise ValueError('Specified option is not in options')
        
        id = self._get_id(option)
        opt = self._options_str[id]
        return self._options.index(opt)
            
    def insert(self, index, option):
        """
        Inserts a new option at the given index.
        Each option must have an unique name.
        
        @index - Index at which to insert the option
        @option - option to append 
        
        @return PopupOption if appending was successful
        
        @raise KeyError if the option already exists
        @raise KeyError if the option is translatable but not in translations
        @raise TypeError if the option is neither PopupOption nor str
        """
        option = self._create_valid_option(option)
        
        self._options.insert(index, option)
        self._options_str[option._name] = option
    
    def keys(self):
        """
        Returns a list of unique PopupOption names used in this Popup
        
        @return list unique names
        """
        return [option._name for option in self._options]
    
    def pop(self, index=-1):
        """
        Removes an option at the given index.
        
        @index - the index at which to remove the option
        
        @return PopupOption if the option at the given index was removed 
                            successfully
        
        @raise IndexError if there are no options or the given index is out of
                          range
        """
        option = self._options.pop(index)
        del self._options_str[option._name]
        return option        
    
    def remove(self, option):
        """
        Removes a given option from the popup.
        
        @option - the option to remove from the popup
        
        @raise ValueError if the option is not present
        """
        if option not in self:
            raise ValueError('Specified option is not in options')
        
        id = self._get_id(option)
        self._options.remove(self._options_str[id])
        del self._options_str[id]
    
    def values(self):
        """
        Returns a list of PopupOption instances used in this Popup
        
        @return list PopupOption instances
        """
        return self._options.copy()
    #
    # Popup stuff
    #
    
    def _format_head(self, args):
        """
        Formats the head of the popup, i.e. the title and description.
        
        Override this function when sub classing and using different options.
        
        @args - formatting/popup arguments
        
        @return list formatted popup lines
        """
        head = []
        if self._title:
            head.append(args['title']['msg'])
        
        if self._description:
            # Insert separator between title and description
            if self._title and self._separator_title:
                head.append(self._separator_title)
            head.append(args['description']['msg'])
        
        if len(head):
            head.append(self._separator_top)
            
        return head
    
    def _format_body(self, body):
        """
        Formats the body of the popup.
        
        Override this function when sub classing and using different options.
        
        @args - formatting/popup arguments
        
        @return list formatted popup lines
        """
        return [data['msg'] for data in body]
    
    def _format_msg(self, args):
        """
        Final formatting and transformation to prepare it for displaying.
        
        Override this function when sub classing and using different options.
        
        @args - formatting/popup arguments
        
        @return list formatted popup lines
        """
        msg = []
        msg += self._format_head(args)
        msg += self._format_body(args['opts'])
        msg.append(self._separator_bottom)
        msg.append('->%s. %s' % (
            OPTION_CLOSE, lang['popup_close'].get_string(args['language'])))
        
        return msg
    
    def _display(self, args):
        """
        Displays this popup using the given arguments
        
        @args - argument dictionary
        """
        # Popup is inactive, don't display
        if not args['active']:
            return
            
        duration = args['duration']
        if duration <= 0:
            duration = -1
        else:
            duration -= time() - args['init_time']
            if duration <= 0:
                args._set_inactive
                
        ShowMenu(
            slots=0b1111111111,  # 0b000000001
            time=duration, 
            message='\n'.join(self._format_msg(args))
        ).send(args['index'])
    
    def _handle_opt(self, opt, playerinfo, language, targs):
        """
        Handles PopupOption to prepare a display string.
        
        @opt      - PopupOption instance
        @kwargs   - Args to pass to opt._prehook
        @language - Language to use when formatting
        @targs    - Translation arguments to use when formatting
        
        @return str the formatted string on success
        @return None on failure (access denied)
        """
        active = True
        if opt._prehook:
            result = opt._prehook(playerinfo, opt, self, language)
            if result == ACCESS_ALLOWED or result is None:
                pass
            elif result == ACCESS_RESTRICTED:
                active = False
            elif result == ACCESS_DENIED:
                return None
        
        rtr = {'msg': '', 'active': False, 'option': opt}
        
        if self._translations and opt._translate:
            id = opt._value if opt._value else opt._name
            
            args = {}
            args.update(targs)
            if opt._translate_callback:
                args.update(opt._translate_callback(
                    playerinfo, opt, self, language))
            
            t = self._translations[id].get_string(language, **args)
            if t == '':
                _log.log(
                    logging.WARN, 
                    'Translation with ID=%s not found or empty' % opt._name)
                rtr['msg'] = opt._name
            else:
                rtr['msg'] = t
        elif opt._value:
            rtr['msg'] = opt._value
        else:
            rtr['msg'] = opt._name
        
        if opt._active and active:
            rtr['active'] = True
            
        return rtr
    
    def send(
            self, playerinfo, targs={}, language=None, sort=None, duration=-1):
        """
        Sends this popup to the given playerinfo instance.
        
        You pass targs, which then will be passed to the internal translation
        method in order to create the appropriately formatted translation.
        
        SORTING:
        
        If you specified a callable function in sort, the parsed options will
        be passed to the sort function.
        
        
        @playerinfo - CPlayerInfo instance
        @targs      - Arguments to pass when formatting translated strings
        @language   - Use this language when formatting; if None, it uses the 
                      player's language
        @sort       - Use this function to sort the arguments
        @duration   - The duration to display this popup for; -1 = infinite
        
        @raise TypeError is the sort function supplied is not callable
        """
        # Bots don't need popups
        if playerinfo.is_fake_client():
            return
        
        if sort and not callable(sort):
            raise TypeError('sort function must be callable')
        
        index = index_from_playerinfo(playerinfo)
        
        if language is not None:
            language = LanguageManager.get_language(language)
        
        # Getting the short-lang might have failed and None is returned
        # Or it was None to begin with
        if language is None:
            language = GameEngine.get_client_convar_value(index, 'cl_language')
            language = LanguageManager.get_language(language)
           
        outopts = []
        for opt in self._options:
            out = self._handle_opt(opt, playerinfo, language, targs)
            if out:
                outopts.append(out)
        
        if sort:
            outopts = sort(outopts, language)
        
        p = PopupManager[playerinfo.get_userid()]
        p['init_time'] = time()
        p['duration'] = duration
        p['popup'] = self
        p['opts'] = outopts
        p['index'] = index
        p['active'] = True
        p['language'] = language
        if self._title:
            p['title'] = self._handle_opt(
                self._title, playerinfo, language, targs)
        else:
            p['title'] = ''
        
        if self._description:
            p['description'] = self._handle_opt(
                self._description, playerinfo, language, targs)
        else:
            p['description'] = ''
        
        self._display(p)
    
    def unsend(self, playerinfo):
        """
        Unsends (hides) this popup for the specified player.
        
        @playerinfo - target's CPlayerInfo instance
        """
        p = PopupManager[playerinfo.get_userid()]
        
        # This popup is not the active popup
        if p['popup'] is not self:
            return
        
        PopupManager.unsend(playerinfo)
    
    def unsend_all(self):
        """
        Unsends (hides) this popup for all players.
        """
        for playerinfo in PlayerGenerator():
            self.unsend(playerinfo)

            
class PopupList(Popup):
    """
    List-style popup which extends the base popup.Popup functionality.
    See popups.Popup for details on the base functionality.
    
    
    PAGED LIST
    
    This popup automatically creates a paged list out of the options given.
    Upon creation, you can specify the amount of options to show per page, and
    while there is no upper limit, you should keep it low, so it doesn't reach
    beyond the users screen or into other UI elements.
    
    Previous/Next options will be automatically generaterated and people can
    navigate though the PopupList easily.
    """
    def __init__(self, options_per_page=OPTIONS_PER_PAGE, **kwargs):
        """
        Creates a new PopupList instance.
        
        @title            - Title of the popup; see Popup.set_title
        @description      - Description of the popup; see Popup.set_description
        @translations     - LangStrings instance to use when translating 
                            options
        @options_per_page - Options per page
        
        @raise TypeError if the title is neither PopupOption, str nor None
        @raise ValueError if options_per_page is lower then 
                          constants.OPTION_MIN
        """
        if options_per_page < OPTION_MIN:
            raise ValueError(
                'You can\'t have less then %s options' % OPTION_MIN)
        self._options_per_page = options_per_page
        self.set_separator(SEPARATOR_BUTTONS, ' ')
        super(PopupList, self).__init__(**kwargs)
    
    def _format_msg(self, args):
        """
        Final formatting and transformation to prepare it for displaying.
        
        Override this function when sub classing and using different options.
        
        @args - formatting/popup arguments
        
        @return list formatted popup lines
        """
        msg = []
        msg += self._format_head(args)
        
        # Pages will start at 0
        l = len(args['opts']) - 1
        pages = l // self._options_per_page
        
        # Fix possible out of bounds pageids
        pid = args['pageid']
        if pid < 0:
            pid = 0
        elif pid > pages:
            pid = pages        

        msg += self._format_body(args['opts'][
            pid * self._options_per_page:(pid + 1) * self._options_per_page])
        msg.append(self._separator_bottom)
        # Don't show on the first page
        if pid > 0:
            msg.append('->%s. %s' % (
                OPTION_PREVIOUS, lang['popup_previous_page'].get_string(
                    args['language'])))
        # Don't show on the last page
        if pid < pages:
            msg.append('->%s. %s' % (
                OPTION_NEXT, lang['popup_next_page'].get_string(
                    args['language'])))
        
        # Add separator
        if pages:
            msg.append(self._separator_close)
            
        msg.append('->%s. %s' % (
            OPTION_CLOSE, lang['popup_close'].get_string(
                args['language'])))
        
        return msg

    
class PopupMenu(PopupList):
    """
    Menu-style popup which extends the popup.PopupList functionality.
    See popups.PopupList and popups.Popup for details on the functionality.
    
    
    PAGED MENU
    
    This popup automatically creates a paged menu out of the options given.
    Upon creation, you can specify the amount of options to show per page, and
    each option will be numbered accordingly from 1 to options_per_page.
    
    Previous/Next options will be automatically generaterated and people can
    navigate though the PopupMenu easily.
    
    
    CALLBACKS
    
    This is the only popup-type that supports callbacks. If a callback is
    specified within the PopupOption, upon selecting an option in the menu
    will call the specified function.
    
    Please note that only works if the specified option is ACTIVE, inactive
    options will NOT trigger callbacks; options may be marked as inactive
    by prehooks or by specifing it explicitly on PopupOption creation.
    """
    def __init__(self, options_per_page=OPTIONS_PER_PAGE, **kwargs):
        """
        Creates a new PopupMenu instance.
        
        @title            - Title of the popup; see Popup.set_title
        @description      - Description of the popup; see Popup.set_description
        @translations     - LangStrings instance to use when translating 
                            options
        @options_per_page - Options per page
        
        @raise TypeError if the title is neither PopupOption, str nor None
        @raise ValueError if options_per_page is:
                          - smaller then constants.OPTION_MIN
                          - larger then constants.OPTION_MAX
        """
        if options_per_page > OPTION_MAX:
            raise ValueError(
                'You can\'t have more then %s options' % OPTION_MAX)
        kwargs['options_per_page'] = options_per_page
        super(PopupMenu, self).__init__(**kwargs)
    
    def _format_body(self, body):
        """
        Formats the body of the popup.
        
        Override this function when sub classing and using different options.
        
        @args - formatting/popup arguments
        
        @return list formatted popup lines
        """
        new = []
        for i in range(0, len(body)):
            if body[i]['active']:
                new.append('->' + str(i + 1) + '. ' + body[i]['msg'])
            else:
                new.append(str(i + 1) + '. ' + body[i]['msg'])
        return new
