"""
Path     addons/source-python/packages/custom/k2tools/pubcvarlib.py
Name     Public Cvar Library
Version  1.00.000
Revision $Rev: 39 $
Author   (C) [#OMEGA] - K2


INFO

General purpose library that makes public variables available.
This also serves as workaround for some issues with settings tags and 
variables.


AGREEMENT

See k2tools_license.txt
"""

__all__ = ['register', 'unregister', 'exists', 'get_cvars']
__version__ = '1.00.000 $Rev: 39 $'
__author__ = '[#OMEGA] - K2'
__date__ = '$Date: 2014-02-09 22:15:20 +0100 (So, 09 Feb 2014) $'

# =============================================================================
# >> Imports
# =============================================================================

# Source Python
import cvar_c

from core import GameEngine
from cvars import ServerVar
from events.manager import EventRegistry

# k2tools Imports
from k2tools.threadlib import GameTimer

# =============================================================================
# >> Globals
# =============================================================================

_cvars = {}
_tags = {}

_sv_tags = ServerVar('sv_tags')

# =============================================================================
# >> Private Functions
# =============================================================================


def _set_tags(string):
    """
    Function to set tags.
    
    sv_tags behaves really weird. Setting though the variable doesn't set it 
    at all, but is required later to get the correct tags value.
    
    If done as if though console it works however, yet the variable can not be
     retrieved properly in-game.
    
    @string - tag to set/add
    """
    _sv_tags.set_string(string)
    GameEngine.server_command('sv_tags "' + string + '";')


def _make_public():
    """
    Try to make all variables in the cache public and also sets the tags .
    """
    curtags = _sv_tags.get_string()
    addtags = []
    for cvar in _cvars:
        if _cvars[cvar]['default']:
            _cvars[cvar]['object'].set_string(_cvars[cvar]['value'])
        else:
            _cvars[cvar]['object'].set_string(
                _cvars[cvar]['object'].get_string())
        for tag in _cvars[cvar]['tags']:
            if tag not in curtags:
                addtags.append(tag)
    if addtags:
        _set_tags(curtags + ',' + ','.join(addtags))


def _server_spawn(ev):
    """
    Event-Listener for server_spawn.
    """
    GameTimer(5, _make_public).start()

# =============================================================================
# >> Public Functions
# =============================================================================


def register(var, value=None, description='', default=False, tags=[]):
    """
    Registers a public variable.
    
    @var         - name of the variable
    @value       - initial value of the variable
    @description - description of the variable
    @default     - if set to true, the variable will revert to the initial 
                   value specified, otherwise the current value will be used
    @tags        - tags to be added to sv_tags
    
    @raise ValueError if the variable exists
    """
    if var in _cvars:
        raise ValueError('Variable "%s" already exists!' % var)
    
    if isinstance(tags, str) or not getattr(tags, '__iter__'):
        tags = [tags, ]
        
    # Make sure the tags are strings and prepare to them if necessary
    curtags = _sv_tags.get_string()
    addtags = []
    for otag in tags:
        tag = str(otag)
        tags[tags.index(otag)] = tag
        if tag not in curtags:
            addtags.append(tag)
    if addtags:
        _set_tags(curtags + ',' + ','.join(addtags))
    
    value = str(value)
    obj = ServerVar(var, value, cvar_c.FCVAR_NOTIFY, description)
    
    _cvars[var] = {
        'value': value, 
        'default': default,
        'object': obj,
        'tags': tags,
    }
    
    # Initial setting of the version var
    GameTimer(1, obj.set_string, (value, )).start()


def unregister(var):
    """
    Unregisters a public variable.
    
    @var - name of the variable
    
    @raise ValueError if the variable does not exist
    """
    if var not in _cvars:
        raise ValueError('Variable "%s" does not exist!' % var)
    
    # Remove tags if necessary
    curtags = _sv_tags.get_string()
    newtags = curtags
    for tag in _cvars[var]['tags']:
        if tag in curtags:
            newtags = newtags.replace(tag + ',', '')
            # Incase it is the last tag
            newtags = newtags.replace(tag, '')
    _set_tags(newtags)
    
    # Take care of the version variable
    o = _cvars[var]['object']
    o.remove_flags(cvar_c.FCVAR_NOTIFY)
    o.add_flags(cvar_c.FCVAR_UNREGISTERED)
    o.set_string('Unloaded')
    
    del _cvars[var]
    

def exists(var):
    """
    Checks whether a specified variable already exists within pubcvarlib
    
    @return True of the specified variable exists
    @return False otherwise
    """
    return var in _cvars


def get_cvars():
    """
    Returns a copy of the internally registered directory for the cvar 
    management.
    
    @return dict a dictionary of cvars (including settings per cvar)
    """
    return _cvars.copy()
    
# =============================================================================
# >> Init
# =============================================================================

EventRegistry.register_for_event('server_spawn', _server_spawn)
