"""
Path     addons/source-python/packages/custom/k2tools/smauth.py
Name     SourceMod Authentication Library
Version  1.00.000
Revision $Rev: 39 $
Author   (C) [#OMEGA] - K2


INFO

Parser for the sourcemod authentication files.


AGREEMENT
 
See k2tools_license.txt
"""

'''
Dev Notes:

Add check for existence of events provided in smauth.ini
PlayerAccess steamid etc might not exist at that point - "Nonetype" error
Add in_groups methods, i.e. InAllGroups/InOneGroups
'''

__k2doc__ = {'Admin': 'instance'}
__all__ = [
    'Admin', 'PlayerAccess', 'flag_to_level', 'level_to_flag', 
    'is_valid_flag', 'is_valid_level', 'is_valid_group']
__version__ = '1.00.000 $Rev: 39 $'
__author__ = '[#OMEGA] - K2'
__date__ = '$Date: 2014-02-09 22:15:20 +0100 (So, 09 Feb 2014) $'

# =============================================================================
# >> Imports
# =============================================================================

# Python
import re

# Source.Python C++
from conversions_c import index_from_playerinfo
from conversions_c import playerinfo_from_index
from conversions_c import playerinfo_from_userid

# Source.Python Python
from paths import GAME_PATH

from core import GameEngine
from cvars import ServerVar
from events.manager import EventRegistry

# k2tools
from .keyvalues import KeyValuesFile
from ._paths import K2TOOLS_CFG_PATH
from ._paths import K2TOOLS_CFGSPEC_PATH
from ._paths import K2TOOLS_LOG_PATH
from .config import load

# =============================================================================
# >> Globals
# =============================================================================

SM_CFG_PATH = GAME_PATH.joinpath('addons', 'sourcemod', 'configs')

config = load(
    K2TOOLS_CFG_PATH.joinpath('smauth.ini'), 
    K2TOOLS_CFGSPEC_PATH.joinpath('smauth-configspec.ini'),
    K2TOOLS_LOG_PATH.joinpath('smauth.log'))

_admin_simple_re = re.compile(r'"([^"]+)"|([^\s]+)')

# =============================================================================
# >> Private Classes
# =============================================================================


class _Admin(object):
    """
    Using this class (or rather the object of this class) should be avoided
    unless you really need it.
    """

    def __init__(self):
        """
        """
        self._parse()
        
    #
    # Private 
    #
        
    def _parse(self):
        self.users = []
        self.groups = {}
        self.levels = {}
        
        # Open the copies with keyvalues
        admin_groups = KeyValuesFile()
        admin_groups.load(SM_CFG_PATH.joinpath('admin_groups.cfg'), typecast=True)
        admin_levels = KeyValuesFile()
        admin_levels.load(SM_CFG_PATH.joinpath('admin_levels.cfg'), typecast=True)
        admins = KeyValuesFile()
        admins.load(SM_CFG_PATH.joinpath('admins.cfg'), typecast=True)
        
        # ################# #
        # admins_simple.ini #
        # ################# #
        with open(SM_CFG_PATH.joinpath('admins_simple.ini')) as adminflatfile:
            lines = adminflatfile.readlines()
        for line in lines:
            # Ingore comments and blank lines
            if not line.startswith('//') and len(line.strip()):
                # Prepare variables
                auth, id = None, None
                options = {
                    'name': None, 
                    'flags': '', 
                    'password': None, 
                    'group': [], 
                    'immunity': 0,
                }
                # Remove comments
                line = line[:line.find('//')]
                # Find the options
                o = re.findall(_admin_simple_re, line)
                if not o:
                    continue
                # Compile a list as the regex contains a tuple
                l = [] 
                for item in o:
                    if item[0]:
                        l.append(item[0])
                    elif item[1]:
                        l.append(item[1])
                # Parse the auth types
                if l[0].startswith('STEAM_'):
                    auth, id = 'steam', l[0]
                elif l[0].startswith('!'):
                    auth, id = 'ip', l[0][1:]
                else:
                    auth, id = 'name', l[0]
                # Parse admin flags
                for item in l[1].split(':'):
                    if item.isdigit():
                        options['immunity'] = item
                    elif item.startswith('@'):
                        options['group'] = item[1:]
                    else:
                        options['flags'] = item
                # Parse the password (incase it exists)
                if len(l) == 3:
                    options['password'] = l[2]
                # Finally add the admin to the list
                self.add_user(auth, id, options)
        
        # ########## #
        # admins.ini #
        # ########## #
        for key in admins['Admins']:
            options = {
                'name': key, 
                'flags': '', 
                'password': None, 
                'group': [], 
                'immunity': 0
            }
            # Loop though the keyvalue pairs
            admin = admins['Admins'][key]
            for subkey in admin:
                value = admin[subkey]
                if subkey == 'group':
                    options[subkey].append(value)
                elif subkey not in ['auth', 'identity']:
                    options[subkey] = value
            # Add the User
            self.add_user(admin['auth'], admin['identity'], options)
        
        # ################ #
        # admin_groups.ini #
        # ################ #
        for groupname in admin_groups['Groups']:
            group = admin_groups['Groups'][groupname]
            keys = group.keys()
            self.add_group(
                groupname, 
                group['flags'] if 'flags' in keys else '', 
                group['immunity'] if 'immunity' in keys else 0)
        
        # ################ #
        # admin_levels.ini #
        # ################ #
        for flag in admin_levels['Levels']['Flags']:
            self.levels[flag] = admin_levels['Levels']['Flags'][flag]
    
    def __get_int(self, var):
        if isinstance(var, (float, int)):
            return var
        elif isinstance(var, str) and var.isdigit():
            return int(var)
        elif var is None:
            return 0
        else:
            raise ValueError('%s can\'t be parsed to an integer value.' % var)
    
    def __add_flags(self, flags1, flags2):
        """
        Adds 2 flag strings together.
        """
        addedflags = ''
        for flag in flags1 + flags2:
            if flag not in addedflags:
                addedflags += flag
        return addedflags
  
    #
    # Public #
    #

    def add_user(self, auth='steam', identity=None, options=None):
        """
        Add user to the cache
        
        @auth     - Authtype
                    - steam : SteamID
                    - name  : Name 
                    - IP    : IP Address
        @identity - Identification string for the authtype
        @options  - Dictionary of options containing the following:
                    - name     : name of the admin
                    - flags    : admin flags
                    - password : user password
                    - group    : list of groups
                    - immunity : immunity level
        """
        # A bit of error checking
        if not auth:
            raise ValueError('Auth type (steam,ip,name) must be provided!')
        elif auth not in ('steam', 'ip', 'name'):
            raise ValueError(
                'Auth type must be either steam, ip or name.\nAuth Type was: '
                ' %s' % auth)
        if not identity:
            raise ValueError('Identity value must be provided!')
            
        if not options:
            options = {
                'name': None, 
                'flags': '', 
                'password': None, 
                'group': [],
                'immunity': 0,
            }
            
        # Make sure immunity is set to a proper value
        if 'immunity' in options:
            options['immunity'] = self.__get_int(options['immunity'])
        # Same for steamid to ensure compability between old and new steam
        # auth method
        if auth == 'steam':
            # Remove the STEAM_X: part from the steamid
            identity = identity[8:]
        
        # Parsing for the group
        if 'group' in options and type(options['group']) == str:
            options['group'] = list(options['group'])
        
        # Add the user
        temp = {'auth': auth, 'identity': identity}
        temp.update(options)
        self.users.append(temp)
    
    def add_group(self, name=None, flags='', immunity=0):
        """
        Add group to the cache. 
        
        @name     - Name of the group
        @flags    - Flags for this group
        @immunity - Immunity level of this group
        """
        # Error Checking
        if not name:
            raise ValueError('Group name must be defined')
        # Check for duplicates
        if name in self.groups:
            raise ValueError(
                'This group already exists. Groups must be unique!')
        # Add the group
        self.groups[name] = {
            'flags': flags, 
            'immunity': self.__get_int(immunity)
        }
    
    def get_user(self, identity, auth='steam'):
        """
        Gets the user object with parsed permissions.
        If the user is not in there, it returns False.
        """
        for user in self.users:
            if not (user['auth'] == auth and user['identity'] == identity):
                continue
            # Adds up immunity and flags for each group the user is in
            for admingroup in ['Default', ] + user['group']:
                for group in self.groups:
                    if group != admingroup:
                        continue
                    if self.groups[group]['immunity'] > user['immunity']:
                        user['immunity'] = self.groups[group]['immunity']
                    user['flags'] = self.__add_flags(
                        user['flags'], self.groups[group]['flags'])
            return user
        return False
    
    def parse(self):
        """
        Clears the cache and re-parses all admins.
        """
        self._parse()
        
# =============================================================================
# >> Public Classes
# =============================================================================


class PlayerAccess(object):
    """
    The purpose of this class is to simplify the access to the data parsed by
    the library. It is possible to use the Admin object directly, however, 
    unless it's not avoidable this class should be preferred.
    """
    def __init__(
            self, userid=None, index=None, player=None, authtype='steam',
            authinfo=None):
        """
        Creates a new player object for the specified index.
        
        You must provide one of userid, index or player to identify the player
        
        @userid   - The userid of the player
        @index    - The index of the player
        @player   - CPlayerInfo instance of the player
        @authtype - The authtype that is used when checking permissions. Can 
                    be either steam, ip or name
        @authinfo - Argument that represents the authentication tag depending
                    on authtype, can be one of the following:
                    - SteamID of player (for authtype 'steam')
                    - IP Address of player (for authtype 'ip')
                    - Name of player (for authtype 'name')
        
        @raise ValueError if the type is invalid (not steam, ip or name) 
        """
        player = self._check_player(userid, index, player)
        self.authtype = authtype.lower()
        if authtype == 'steam':
            if authinfo:
                if authinfo.startswith('STEAM_'):
                    authinfo = authinfo[8:]
                self.authinfo = authinfo
            else:
                self.authinfo = player.get_networkid_string()[8:]
        elif authtype == 'ip':
            if authinfo:
                self.authinfo = authinfo
            else:
                if not index:
                    index = index_from_playerinfo(player)
                socket = GameEngine.get_player_net_info(index).get_address()
                self.authinfo = socket.split(':')[0] if socket else socket
        elif authtype == 'name':
            self.authinfo = authinfo if authinfo else player.get_name()
        else:
            raise ValueError(
                'Auth type must be either steam, ip or name.\n Auth Type was: '
                '%s' % auth)
        # Get the userinfo
        self.user = Admin.get_user(self.authinfo, self.authtype)
    
    def _check_player(self, userid=None, index=None, player=None):
        """
        Try to get the CPlayerInfo object from a given authentication method.
        
        @userid   - The userid of the player
        @index    - The index of the player
        @player   - CPlayerInfo instance of the player
        
        @return CPlayerInfo the object on success
        
        @raise ValueError if no valid keywords were specified
        """
        if userid:
            player = playerinfo_from_userid(userid)
        elif index:
            player = playerinfo_from_index(index)
        elif not player:
            raise ValueError(
                'One of the keywords "userid", "index" or "player" must be '
                'specified')
       
        return player
    
    def is_user(self):
        """
        Checks whether the user is stored in the database
        
        @return True if the user with the specified auth is stored in admin
                files
        @return False otherwise 
        """
        return bool(self.user)
    
    def is_admin(self):
        """
        Checks whether the user has basic admin access (b) or not.
        
        @return True if the user has the basic admin (b) flag
        @return False if the user does not exist or if the flag can not be
                found.
        """
        if self.is_user():
            for flag in 'bz':
                if flag in self.user['flags']:
                    return True
        return False
    
    def is_root(self):
        """
        Checks whether the user has root admin access (z) or not.
        
        @return True if the user has root access.
        @return False otherwise.
        """
        return self.has_flag()
    
    def has_flag(self, adminflag=''):
        """
        This functions checks whether the user has the specified flag. It also
        checks against whether he has root access and thus is able to use the
        function without having the flag itself.
        
        @adminflag    - A single character which represents the admin flag.
        
        @return True if the user has the specified flag or has root access.
        @return False if the user does not exist or if the flag can not be 
                      found.
        
        @raise TypeError if adminflag is not a string.
        """
        if not isinstance(adminflag, str):
            raise TypeError(
                "Invalid type for adminflags specified. Type received was"
                "%s." % type(adminflags))
        
        if self.is_user():
            for flag in adminflag + 'z':
                if flag in self.user['flags']:
                    return True
        return False
    
    def has_flags(self, adminflags=()):
        """
        This functions checks whether the user has all specified flag. It also
        checks against whether he has root access and thus is able to use the
        function without having the flag itself.
        
        @adminflags    - Any iterable type (dict, list, tuple or strings)
        
        @return True if the user has all the specified flags (or root flag).
        @return False if the has one of the specified flags or if the flag can
                      not be found.
        
        @raise TypeError if the type is invalid.
        """
        if not isinstance(adminflags, (dict, list, tuple, str)):
            raise TypeError(
                "Invalid type for adminflags specified. Type received was "
                "%s." % type(adminflags))
        
        if not self.is_user():
            return False
            
        if 'z' in self.user['flags']:
            return True
            
        for flag in adminflags:
            if flag not in self.user['flags']:
                return False
        
        return True
    
    def get_immunity(self):
        """
        This function is used to receive the immunity level of the user. It
        will always return the highest applicable immunity.
        
        @return int the immunity value of the user
        @return False otherwise
        """
        return self.user['immunity'] if self.is_user() else False
    
    def in_group(self, group):
        """
        This functions checks whether the user is in the specified admin group.
        This function is case-sensitive.
        
        @group    - The admin group
        
        @return True if the user is in the specified group.
        @return False if the user does not exist or if the group can not be 
                      found.
        """
        return self.is_user() and group in self.user['group']
    
    def can_target(self, *a, userid=None, index=None, player=None):
        """
        This functions checks whether the user can target the player with the 
        specified index with an admin command. Depending on sm_immunity_mode 
        the according value may vary. See SourceMod documentation for more 
        info on what sm_immunity_mode changes.
        
        Only one of the keyword arguments is required to identify the target 
        player.
        
        @userid - The userid of the targeted player
        @index  - The index of the targeted player
        @player - CPlayerInfo instance of the targeted player 
        
        @return True if the user can target the specified index with an admin 
                     command.
        @return False if the user does not exist or if the index can not be 
                      targeted by the user.
        """
        if not self.is_user():
            return False
       
        immunity_mode = ServerVar('sm_immunity_mode').get_int()
        # If 0 immunity levels are ingored anyway...
        if not immunity_mode:
            return True
        
        # Root admins can target anyone, no matter the immunity.
        if self.is_root():
            return True
        
        for authtype in ('steam', 'ip', 'name'):
            target = PlayerAccess(userid, index, player, authtype)
            if not target.is_user():
                return True
            if immunity_mode == 1:
                return self.get_immunity() >= target.get_immunity()
            elif immunity_mode == 2:
                return self.get_immunity() > target.get_immunity()
            elif immunity_mode == 3:
                if self.get_immunity() == 0 and target.get_immunity() == 0:
                    return True
                else:
                    return self.get_immunity() > target.get_immunity()
        return True

# =============================================================================
# >> Private Functions
# =============================================================================


def _register_auto_refresher():
    event = config['General']['AutoRefreshEvent']
    if len(event):
        EventRegistry.register_for_event(event, _auto_refresh_event)


def _auto_refresh_event(ev):
    Admin.parse()

# =============================================================================
# >> Public Functions
# =============================================================================


def flag_to_level(flag):
    """
    Converts the flag to the according sourcemod access level.
    
    @flag    - the flag that is supposed to be converted to the corresponding 
               level
    
    @raise ValueError if the flag doesn't exist.
    """
    for level in Admin.levels:
        if Admin.levels[level] == flag:
            return level
    raise ValueError('Flag %s does not exist!' % flag)


def level_to_flag(level):
    """
    Converts the sourcemod access level to the according sourcemod flag.
    This is case-sensitive.
    
    @level    - the name of the admin level that is supposed to be converted 
                to the corresponding flag
    
    @raise ValueError if the level doesn't exist.
    """
    if is_valid_level(level):
        return Admin.levels[level]
    else:
        raise ValueError('Access level %s does not exist' % level)


def is_valid_flag(flag):
    """
    This function can be used to check whether the specified flag is valid. 
    This function is case-sensitive.
    
    @flag    - the flag you want to validate
    
    @return True if the specified flag is valid and exists.
    @return False otherwise 
    """
    for level in Admin.levels:
        if Admin.levels[level] == flag:
            return True
    return False

    
def is_valid_level(level):
    """
    This function can be used to check whether the specified admin level is 
    valid. This function is case-sensitive.
    
    @level    - the level you want to validate
    
    @return True if the specified level is valid and exists.
    @return False otherwise
    """
    return level in Admin.levels
    

def is_valid_group(group):
    """
    This function can be used to check whether the specified admin group is
    valid. This function is case-sensitive.
    
    @group    - the name of the group you want to validate
    
    @return True if the specified group is valid and exists.
    @return False otherwise
    """
    return group in Admin.groups

# =============================================================================
# >> Init
# =============================================================================
Admin = _Admin()
_register_auto_refresher()
