"""
Path     addons/source-python/packages/custom/k2tools/string.py
Name     String-related functions and methods
Version  1.00.000
Revision $Rev: 39 $
Author   (C) [#OMEGA] - K2


INFO

Miscellaneous functions and classes for strings.


AGREEMENT

See k2tools_license.txt


TODO

- Possibly use own string search function for parsing, right now I suppose it 
searches the string quite often, depending on the tags registered. But for now
speed should be sufficient.
"""

__k2doc__ = {'TagTypeManager': 'instance'}
__all__ = ['TagTypeManager', 'TagParser', 'rfill', 'lfill', 'double_fill']
__version__ = '1.00.000 $Rev: 39 $'
__author__ = '[#OMEGA] - K2'
__date__ = '$Date: 2014-02-09 22:15:20 +0100 (So, 09 Feb 2014) $'

# =============================================================================
# >> Imports
# =============================================================================

# Python
import re
from time import strftime

# Source.Python
from conversions_c import index_from_playerinfo

# Source.Python
from core import GameEngine
from cvars import ServerVar
from translations.manager import LanguageManager

# K2tools
from .data import get_current_map
from .data import server_ip
from .data import server_port

# =============================================================================
# >> Globals
# =============================================================================

_cvar_re = re.compile(r'\{CVAR:[^\}]*\}')

# =============================================================================
# >> Classes
# =============================================================================


class _TagTypeManager(dict):
    """
    Dictionary-style Manager that holds the registered tags. It comes with 
    some default tags and the appropriate functions.
    
    When registering your own Tags, simply add them to the manager. Tag 
    functions should accept two parameters, string and **kwargs.
    With kwargs you can access additional parameters that were passed to the
    function.
    """
    _none = 'None'
    _timeformat = '%H:%M:%S'
    _dateformat = '%d.%m.%Y'
    
    def __init__(self):
        super(_TagTypeManager, self).__init__()
        self['NAME'] = self.name
        self['STEAMID'] = self.steamid
        self['USERID'] = self.userid
        self['CLIENT_IP'] = self.client_ip
        self['CLIENT_LANG'] = self.client_lang
        self['CLIENT_LANG_CODE'] = self.client_lang_code
        self['SERVER_IP'] = self.server_ip
        self['SERVER_PORT'] = self.server_port
        self['CURRENT_MAP'] = self.current_map
        self['DATE'] = self.date
        self['TIME'] = self.time
        self['CVAR'] = self.cvar
        
    def _has_key(self, kwargs, key):
        return key in kwargs and kwargs[key]
   
    def name(self, string, **kwargs):
        """
        Replaces {NAME} with the name of the player.
        
        @string     - string to perform the replacement with
        @playerinfo - CPlayerInfo instance of the player
        
        @return str formatted string
        """
        if self._has_key(kwargs, 'playerinfo'):
            name = kwargs['playerinfo'].get_name()
        else:
            name = self._none
        return string.replace('{NAME}', name)
   
    def steamid(self, string, **kwargs):
        """
        Replaces {STEAMID} with the SteamID of the player.
        
        @string     - string to perform the replacement with
        @playerinfo - CPlayerInfo instance of the player
        
        @return str formatted string
        """
        if self._has_key(kwargs, 'playerinfo'):
            steamid = kwargs['playerinfo'].get_networkid_string()
        else:
            steamid = self._none
        return string.replace('{STEAMID}', steamid)
    
    def userid(self, string, **kwargs):
        """
        Replaces {USERID} with the userid of the player.
        
        @string     - string to perform the replacement with
        @playerinfo - CPlayerInfo instance of the player
        
        @return str formatted string
        """
        if self._has_key(kwargs, 'playerinfo'):
            userid = kwargs['playerinfo'].get_userid()
            userid = str(userid)
        else:
            userid = self._none
        return string.replace('{USERID}', userid)
    
    def client_ip(self, string, **kwargs):
        """
        Replaces {CLIENT_IP} with the IP Address of the player.
        
        @string     - string to perform the replacement with
        @playerinfo - CPlayerInfo instance of the player
        
        @return str formatted string
        """
        if self._has_key(kwargs, 'playerinfo') and not \
                kwargs['playerinfo'].is_fake_client():
            index = index_from_playerinfo(kwargs['playerinfo'])
            socket = GameEngine.get_player_net_info(index).get_address()
            addr = socket.split(':')[0] if socket else socket
        else:
            addr = self._none
        return string.replace('{CLIENT_IP}', addr)
    
    def client_lang(self, string, **kwargs):
        """
        Replaces {CLIENT_LANG} with the language of the player.
        
        @string     - string to perform the replacement with
        @playerinfo - CPlayerInfo instance of the player
        
        @return str formatted string
        """
        if self._has_key(kwargs, 'playerinfo'):
            index = index_from_playerinfo(kwargs['playerinfo'])
            lang = GameEngine.get_client_convar_value(index, 'cl_language')
            if not lang:
                lang = self._none
        else:
            lang = self._none
        return string.replace('{CLIENT_LANG}', lang)
    
    def client_lang_code(self, string, **kwargs):
        """
        Replaces {CLIENT_LANG_CODE} with the language code id of the player.
        
        @string     - string to perform the replacement with
        @playerinfo - CPlayerInfo instance of the player
        
        @return str formatted string
        """
        if self._has_key(kwargs, 'playerinfo'):
            index = index_from_playerinfo(kwargs['playerinfo'])
            lang = GameEngine.get_client_convar_value(index, 'cl_language')
            langid = LanguageManager.get_language(lang)
            if langid is None:
                langid = self._none
        else:
            langid = self._none
        return string.replace('{CLIENT_LANG_CODE}', langid)
    
    def server_ip(self, string, **kwargs):
        """
        Replaces {SERVER_IP} with the server's ip address.
        
        @string - string to perform the replacement with
        
        @return str formatted string
        """
        return string.replace('{SERVER_IP}', server_ip)
    
    def server_port(self, string, **kwargs):
        """
        Replaces {SERVER_PORT} with the server's port.
        
        @string - string to perform the replacement with
        
        @return str formatted string
        """
        return string.replace('{SERVER_PORT}', str(server_port))
    
    def current_map(self, string, **kwargs):
        """
        Replaces {CURRENT_MAP} with the server's current map.
        
        @string - string to perform the replacement with
        
        @return str formatted string
        """
        return string.replace('{CURRENT_MAP}', get_current_map())
    
    def date(self, string, **kwargs):
        """
        Replaces {DATE} with the current date in the given format.
        
        @string - string to perform the replacement with
        @format - python format used to format the string
        
        @return str formatted string
        """
        if self._has_key(kwargs, 'format'):
            format = kwargs['format']
        else:
            format = self._dateformat
        return string.replace('{DATE}', strftime(format))
    
    def time(self, string, **kwargs):
        """
        Replaces {TIME} with the current time in the given format.
        
        @string - string to perform the replacement with
        @format - python format used to format the string
        
        @return str formatted string
        """
        if self._has_key(kwargs, 'format'):
            format = kwargs['format']
        else:
            format = self._timeformat
        return string.replace('{TIME}', strftime(format))
    
    def cvar(self, string, **kwargs):
        """
        Replaces {CVAR:<name>} with the current value of the cvar with <name>
        
        @string - string to perform the replacement with
        
        @return str formatted string
        """
        cvars = re.findall(_cvar_re, string)
        for cvar in cvars:
            string = string.replace(cvar, ServerVar(cvar[6:-1]).get_string())
        return string


class TagParser(object):
    """
    Class for parsing tags out a string.
    """
    def __init__(self, tags=None):
        """
        Creates a new TagParser instance. It allows you to parse only tags you
        specified not all, which allows for faster parsing, as less checks for
        tags are made, only the ones necessary.
        
        Any tags supplied must be valid and registered with TagTypeManager
        
        @tags - The tags you want to be parsed. Leave empty for all tags.
        
        @raise KeyError if the tag type is not registered
        """
        self.tags = []
        if tags:
            for tag in tags:
                if tag not in TagTypeManager:
                    raise KeyError('This tag-type is not registered')
                self.tags.append(tag)
        else:
            self.tags = TagTypeManager.keys()
    
    def parse(self, string, **kwargs):
        """
        Parses the tags for a given string with the given keyword arguments.
        Special keyword arguments may be required by some tags such as player 
        based tags (they require 'playerinfo' to be passed).
        
        @string - the string to be parsed
        
        @return str the formatted/parsed string
        """
        for tag in self.tags:
            string = TagTypeManager[tag](string, **kwargs)
        return string
        
# =============================================================================
# >> Functions
# =============================================================================

    
def rfill(string, length, char=' '):
    """
    Fills the right side of a string until it reaches length with the 
    specified character.
    
    @string - string to fill
    @length - length to fill up to
    @char   - character to use to fill
    
    @return str filled string
    """
    return string + char * (length - len(string))

    
def lfill(string, length, char=' '):
    """
    Fills the left side of a string until it reaches length with the specified
    character.
    
    @string - string to fill
    @length - length to fill up to
    @char   - character to use to fill
    
    @return str filled string
    """
    return char * (length - len(string)) + string


def double_fill(string, length, char=' '):
    """
    Fills both sides of a string until it reaches the specified length with 
    the specified character.
    
    @string - string to fill
    @length - length to fill up to
    @char   - character to use to fill
    
    @return str filled string
    """
    chars = length - len(string)
    half = chars // 2
    string = char * half + string + char * half
    if chars >= 0 and chars % 2:
        string += char
    return string
    
# =============================================================================
# >> Globals/Init
# =============================================================================

TagTypeManager = _TagTypeManager()
