﻿"""
Path     addons/source-python/packages/custom/k2tools/threaddb.py
Name     Threaded DBPAI helper libary
Version  1.00.000
Revision $Rev: 8 $
Author   (C) [#OMEGA] - K2


IMPORTANT

This requires a DBAPI 2.0 compatible driver.
Specifically tested ONLY with pymysql.


INFO

This library is intended to provide some more suitable way of using Database 
queries on a srcds gameserver without interfering with the main server thread.

Usually DBAPI queries halt the server in one way or another, as such, this 
library will execute them in threaded fashion. Please note this was only 
tested with pymysql and as such may not work correctly with other DBAPI 
drivers, even though the general support is there.

In addition to the threads, errors that cause connection loss or otherwise are
handled and the plugin will try to reconnect to the DB if connection was lost.

If you encounter any errors, and they are not handled (and preventing the 
plugin from running) please contact me and I'll see what I can do.

You can find information about MySQL error codes here:
http://dev.mysql.com/doc/refman/5.5/en/error-messages-server.html
http://dev.mysql.com/doc/refman/5.5/en/error-messages-client.html


AGREEMENT
 
See k2tools_license.txt
"""

'''

Todo / Ideas:

-mysql queue: multiple thread support/worker thread per queue
'''

__k2doc__ = {'DBAPIManager': 'instance'}
__all__ = ['DBAPI', 'CBWrapper', 'DBAPIManager']
__version__ = '1.00.000 $Rev: 8 $'
__author__ = '[#OMEGA] - K2'
__date__ = '$Date: 2012-12-26 16:34:11 +0100 (Mi, 26 Dez 2012) $'

# =============================================================================
# >> Imports
# =============================================================================

# Custom Imports
from .threadlib import game_thread

# Python Imports
import time
import socket
import encodings  # In order to fix: LookupError
import threading
import logging

# =============================================================================
# >> Globals
# =============================================================================

_drivers = {}
_lastid = 0
_logger = logging.getLogger('k2tools.threaddb')

# =============================================================================
# >> Private Functions
# =============================================================================


def _load_driver(name):
    """
    Loads specified driver.
    
    @name - Name of the driver
    
    @raise ValueError if no driver name was given
    @raise ImportError if the driver is invalid
    """
    if not name:
        raise ValueError('Driver name must be specified')
    
    if name in _drivers:
        return _drivers[name]
        
    driver = __import__(name)
    
    if not hasattr(driver, 'apilevel') or driver.apilevel != '2.0':
        raise ImportError('Invalid DPAPI 2.0 driver!')
        
    # Threads may not share the module.
    if driver.threadsafety == 0:
        return driver
    # 1 Threads may share the module, but not connections.
    # 2 Threads may share the module and connections.
    # 3 Threads may share the module, connections and cursors.
    elif driver.threadsafety >= 1:
        _drivers[name] = driver
        return driver
   

# =============================================================================
# >> Private Classes
# =============================================================================


class _ConsoleHandler(logging.Handler):
    def emit(self, record):
        game_thread(print, self.format(record))


class _DBAPIManager(threading.Thread):
    """
    Don't call any of the thread methods. It may cause unexpected behaviour.
    """
    def __init__(self):
        self._objects = {}
        super(_DBAPIManager, self).__init__()
    
    def create_object(self, name, *args, **kwargs):
        """
        Creates a named DBAPI object.
        
        Creating a named DBAPI object is useful to retrive it upon script 
        reloading and if running 2 sperate threads is an issue. For example, 
        upon reload your script might rely on properly writing the old data to
        the database, however if 2 threads are running you might end up 
        writing and getting data from the db at the same time.
        
        To circumvent this issue you can a named object; at a later point you 
        can retrive the object as long it is still alive and executing SQL 
        queries.
        
        @name   - internal name of the DBAPI object, must be unique
        
        @raise KeyError if the object already exists
        """
        if name in self._objects:
            raise KeyError("Key %s already exists!" % name)
            
        kwargs['name'] = name
        self._objects[name] = {'object': DBAPI(*args, **kwargs), 'ref': 1}
        self._objects[name]['object']._logger.log(
            logging.INFO, "DBAPI object %s created", name)
        return self._objects[name]['object']

    def delete_object(self, name):
        """
        Reduces the refcount to the specified DBAPI object by one.
        If the refcount hits 0 it will be deleleted automatically once all
        commands are executed and the connection will be seperated.
        
        @name - internal name of the DBAPI object
        
        @raise KeyError if the object is not found
        """
        if name not in self._objects:
            raise KeyError("Key %s does not exists!" % name)
        self._objects[name]['ref'] -= 1
        self._objects[name]['object']._logger.log(
            logging.INFO, "%s's refcount decreased to %s", name, 
            self._objects[name]['ref'])
    
    def get_object(self, name):
        """
        Increases the refcount to the specified DBAPI object by one.
        
        @name - internal name of the DBAPI object
        
        @return DBAPI the sql object if found
        @return False if the sql object is dead
        @return False otherwise
        """
        if name not in self._objects:
            return False
        
        obj = self._objects[name]['object']
        if (self._objects[name]['ref'] < 1):
            if not obj._active:
                if len(obj.queue) < 1:
                    return False
                obj._active = True
        self._objects[name]['ref'] += 1
        obj._logger.log(
            logging.INFO, "%s's refcount increased to %s", name, 
            self._objects[name]['ref'])
        return obj
    
    def run(self):
        """
        DO NOT CALL THIS METHOD!
        
        Removes dead MySQL objects.
        Sets MySQL objects inactive / disconnects them if ref count reaches 0.
        """
        delnames = []
        for name in self._objects:
            if self._objects[name]['ref'] < 1:
                if self._objects[name]['object'].is_alive():
                    delnames.append(name)
                elif self._objects[name]['object'].active:
                    self._objects[name]['object'].disconnect()
        for name in delnames:
            self._objects[name]['object']._logger.log(
                logging.INFO, "DPAPI object %s deleted", name)
            del self._objects[name]
        time.sleep(60)

# =============================================================================
# >> Public Classes
# =============================================================================


class CBWrapper(object):
    """
    Wrapper class just in case DB callback occurs after the script is unloaded.
    To use just create an object of this class and set unloaded to True upon
     unload.
    
    @unloaded - If True, don't execute callbacks any more
    """
    class _CB(object): 
        def __init__(self, parent, cb):
            self.cb = cb
            self.parent = parent
       
        def __call__(self, *args, **kwargs):
            if self.parent.unloaded:
                return
            self.cb(*args, **kwargs)
    
    def __init__(self):
        self.unloaded = False
    
    def cb(self, cb):
        """
        Use this function to wrap a class.
        
        @cb - The callback you want to wrap
        
        @return CBWrapper._CB Wrapped function 
        """
        return CBWrapper._CB(self, cb)


class DBAPI(threading.Thread):
    def __init__(
            self, driver=None, sleep=1, traise=False, loglevel=logging.ERROR, 
            logfile=None, on_connect_queries=(), debug=1, *args, **kwargs):
        """
        Creates a threaded DBAPI helper class.

        @driver             - DBAPI Driver to use. (Default: auto)
                              Supported drivers (others might work too):
                              sqlite3
                              pymysql
        @sleep              - How often it should check whether new queries 
                              were added to the queue in seconds.
        @traise             - If false, don't raise exceptions in threads but 
                              send the error to the console.
        @loglevel           - Errorlevel to use
        @on_connect_queries - Iterable of queries to execute when connected to
                              the server
        @debug              - Extended error checking

        @raise ImportError if no valid driver is detected.
        """
        self._driver = _load_driver(driver)
        self._debug = debug
        
        self.on_connect_queries = on_connect_queries
        
        # Threading stuff
        self._sleep = sleep
        self._traise = traise
        
        # Misc
        self._args = ()
        self._kwargs = {}
        self._connected = False
        # Thread stuff
        self._active = True
        self._queue = []
        
        global _lastid
        
        if 'name' in kwargs:
            kwargs['name'] = kwargs['name'] + '-' + str(_lastid)
        else:
            kwargs['name'] = 'threaddb-' + str(_lastid)
        
        _lastid += 1
        
        super(DBAPI, self).__init__(*args, **kwargs)
        
        self._logger = logging.getLogger('k2tools.threaddb.' + self.getName())
        self._hfile = None
        if logfile:
            self.change_logfile(logfile)
        
        self._hcon = _ConsoleHandler()
        self._hcon.setFormatter(logging.Formatter(
            '[k2tools.threaddb] [Thread %(threadName)s] [%(levelname)s] '
            '%(message)s'))
        self._logger.addHandler(self._hcon)
        self._logger.setLevel(loglevel)
        
        self.start()
    
    def _connect(self, *args, **kwargs):
        if len(args) or len(kwargs):
            self._args = args
            self._kwargs = kwargs
        else:
            args = self._args
            kwargs = self._kwargs
        try:
            self._con = self._driver.connect(*args, **kwargs)
        except self._driver.OperationalError as error:
            # Handling for failure to connect to MySQL server
            if error.args[0] == 2003:
                self._logger.log(logging.ERROR, error.args[1])
                self._connected = False
                self._connect()
            else:
                raise self._driver.OperationalError(error.args)
        else:
            self._connected = True
            self._cur = self._con.cursor()
            for query in self.on_connect_queries:
                self._execute(query, cb=False)
    
    def _execute(self, query, args=(), cb=True):
        try:
            r = self._cur.execute(query, args)
        except (self._driver.Error, self._driver.OperationalError, 
                socket.error) as error:
            # MySQL Errors:
            # 2006  - "MySQL Server has gone away"
            # 2013  - Lost connection to MySQL server during query
            #
            # Socket Errors:
            # 32    - Broken Pipe
            # 10053 - Software caused connection abort (pymysql.err.Error for
            #         some reason)
            # 10054 - Connection Reset by peer
            # 10060 - Connection timout
            # 10061 - Connection refused
            
            # pymysql.err.Error contains a (socket.error (id, message))
            if error.args[0] == socket.error:
                error = error.args[0]
            if error.args[0] in (32, 2006, 2013, 10053, 10054, 10060, 10061):
                self._logger.log(logging.ERROR, error.args[1], ())
                self._connect()
                self._execute(query, args, cb)
            else:
                raise
    
    def run(self):
        """
        DO NOT CALL THIS FUNCTION!!!
        
        This is the main loop that executes the queries running inside a 
        thread.
        """
        while self._active:
            while len(self._queue):
                try:
                    self._execute(
                        self._queue[0]['query'], self._queue[0]['args'])
                except Exception as error:
                    # This is just for outputting extra information for the
                    # programmer
                    self._logger.log(
                        logging.ERROR, 'SQL Query: %s', 
                        self._queue[0]['query'])
                    self._logger.log(
                        logging.ERROR, 'SQL Args: %s', self._queue[0]['args'])
                    if self._traise:
                        raise error.__class__(error.args)
                    else:
                        self._logger.log(
                            logging.ERROR, "%s: %s", error.__class__.__name__,
                            error.args)
                        self._queue.pop(0)
                        continue
                self._logger.log(
                    logging.INFO, 'SQL Query executed. Remaining queries: %s',
                    len(self._queue))
                self._logger.log(
                    logging.DEBUG, 'SQL Query: %s', self._queue[0]['query'])
                self._logger.log(
                    logging.DEBUG, 'SQL Args: %s', self._queue[0]['args'])
                if self._queue[0]['callback']:
                    game_thread(
                        self._queue[0]['callback'], 
                        *((self._cur.fetchall(),) + 
                            self._queue[0]['callback_params']), 
                        **self._queue[0]['callback_kwparams'])
                self._queue.pop(0)
            time.sleep(self._sleep)
        self._cur.close()
        self._con.close()
        if self._hfile:
            self._hfile.close()
            
    #
    # Intended for public use
    #
    def get_driver(self):
        """
        Returns the currently used driver for this DBAPI object.
        
        @return module driver module
        """
        return self._driver
    
    def get_cursor(self):
        """
        Returns the currently used cursor for this DBAPI object.
        
        @return Cursor current cursor object
        """
        return self._cur
    
    def get_connection(self):
        """
        Returns the currently used connection for this DBAPI object.
        
        @return Connection current connection object
        """
        return self._con
    
    def change_logfile(self, newfile):
        """
        Changes the current logfile to a new one. This will make sure the old 
        log file is properly closed.
        
        @newfile - the new logfile to use; use None to disable 
        """
        if self._hfile:
            self._logger.removeHandler(self._hfile)
            self._hfile.close()
        if newfile:
            new = logging.FileHandler(newfile)
            new.setFormatter(logging.Formatter(
                '%(asctime)s - %(levelname)s - %(message)s'))
            self._logger.addHandler(new)
        else:
            new = None
        self._hfile = new
    
    def connect(self, *args, **kwargs):
        """
        Connects to the MySQL server. 

        This function will directly pass any arguments or keyword arguments to
        the driver's connect function.
        See help on <driver>.connect (i.e. pymysql.connect) for more info.
        
        @args   - positional arguments to pass to the driver's connect function
        @kwargs - keyword arguments to pass to the driver's connect function
        """
        self._connect(*args, **kwargs)
    
    def execute(self, query, *args):
        """
        Executes a query. 

        This is NOT threaded.

        @query    - The SQL query to execute
        @args     - Arguments to insert into the query
        """
        self._execute(query, args, False)
    
    def texecute(self, query, args=(), callback=None, params=(), kwparams={}):
        """
        Appends a query to the query queue. This is threaded and will not halt
        the server. 
        
        Does extensive type checks with debug enabled. Otherwise make sure the
        params are of the correct type or the queue might break.

        @query       - The SQL query to execute
        @args        - Arguments to insert into the query
        @callback    - Function to be called once the query was successful.
                       The function must have at least 1 paramater in which the
                       results of the query will be stored (as received by
                       fetchall()).
        @params      - Additional function parameters
        @kwparams    - Additional keyword function parameters
        
        @raise TypeError (debug) if args, params or kwparams is of invalid type
        """
        if self._debug:
            if not isinstance(args, (tuple, list)):  # must correspond to type
                raise TypeError("Arguments must be a sequence")
            if not isinstance(params, (tuple, list)):
                raise TypeError("Arguments must be a sequence")
            if not isinstance(kwparams, dict):
                raise TypeError("Keyword arguments must be a dictionary")
        self._queue.append({
            'query': query,
            'args': args,
            'callback': callback,
            'callback_params': params,
            'callback_kwparams': kwparams,
        })
    
    def disconnect(self):
        """
        Disconnects from the database. 

        This will continue executing any remaining queries in the queue before
        actually closing the connection.
        """
        self._active = False

# =============================================================================
# >> Init
# =============================================================================
        
DBAPIManager = _DBAPIManager()
DBAPIManager.start()
