"""
Path     addons/source-python/packages/custom/k2tools/threadlib.py
Name     Threading Helper Library
Version  1.00.000
Revision $Rev: 39 $
Author   (C) [#OMEGA] - K2


INFO

This is a helper library to make threading easier in Source Python. It provides
a few general purpose methods which allow you to avoid trouble with threading.


AGREEMENT

See k2tools_license.txt
"""

__all__ = ['game_thread', 'tick_delay', 'GameTimer', 'ThreadManager']
__version__ = '1.00.000 $Rev: 39 $'
__author__ = '[#OMEGA] - K2'
__date__ = '$Date: 2014-02-09 22:15:20 +0100 (So, 09 Feb 2014) $'

# =============================================================================
# >> Imports
# =============================================================================

from sys import exc_info
from threading import Thread, Timer
from queue import Queue

# Source Python
from listener_c import TickListenerManager

from excepthooks import ExceptHooks
 
# =============================================================================
# >> Classes
# =============================================================================


class ThreadManager(list):
    """
    Helper class that handles Thread objects and can remove them easily on 
    plugin unload.
    """
    def _remove_timer(self, timer):
        self.pop(self.index(timer))
    
    def new_thread(self, *args, **kwargs):
        """
        Creates a new Thread object, appends it to the ThreadManager
        instance and returns it.
        
        Also see Thread.__init__ 
        
        @return Thread new Thread object
        """
        self.append(Thread(*args, **kwargs))
        return self[-1]
    
    def new_timer(self, *args, **kwargs):
        """
        Creates a new Timer object, appends it to the ThreadManager
        instance and returns it.
        
        Object will removed from ThreadManager when the timer is executed
        or canceled.
        
        Also see Timer.__init__
        
        @return Timer new Timer object
        """
        t = _Timer(*args, **kwargs)
        t._tm_remove = self._remove_timer
        self.append(t)
        return t
    
    def new_game_timer(self, *args, **kwargs):
        """
        Creates a new GameTimer object, appends it to the ThreadManager
        instance and returns it.
        
        Object will removed from ThreadManager when the timer is executed
        or canceled.
        
        Also see Timer.__init__ 
        Also see GameTimer
        
        @return GameTimer new GameTimer object
        """
        t = _GameTimer(*args, **kwargs)
        t._tm_remove = self._remove_timer
        self.append(t)
        return t
    
    def unload(self):
        """
        Call this method when you unload your script.
        
        This will cancel all timers created though this class and will make
        any threads join the main thread (stop until all threads are processed)
        """
        for item in self:
            if isinstance(item, Timer):
                item.cancel()
            else:
                # Block until the thread is done executing
                item.join()

                
class _TimerHelper(object):
    def cancel(self):
        super(self.__class__, self).cancel()
        self._tm_remove(self)
    
    def run(self):
        try:
            super(self.__class__, self).run()
        finally:
            self._tm_remove(self)

 
class _Timer(Timer, _TimerHelper):
    pass

                      
class GameTimer(Timer):
    """
    This timer ensures that the code will be executed in a game frame. 
    Otherwise it just behaves like a normal Timer.
    """
    def run(self):
        """
        DO NOT CALL THIS FUNCTION!!!
        
        Override to the default run method to ensure the Timer works correctly
        """
        self.finished.wait(self.interval)
        if not self.finished.is_set():
            game_thread(self.function, *self.args, **self.kwargs)
        self.finished.set()

 
class _GameTimer(GameTimer, _TimerHelper):
    pass
 
# =============================================================================
# >> Functions
# =============================================================================

 
def _tick_listener():
    while(_queue.qsize() != 0):
        data = _queue.get(False)
        # The method may cause problems causing other methods not to execute
        # Need to handle this
        try:
            data['cb'](*data['args'], **data['kwargs'])
        except:
            # Probably should be replaced by something more suitable
            ExceptHooks.print_exception(*exc_info())
        finally:
            _queue.task_done()
    
    global _tick
    _tick += 1
    
    for item in _tick_queue:
        if item['end'] <= _tick:
            try:
                item['cb'](*item['args'], **item['kwargs'])
            except:
                ExceptHooks.print_exception(*exc_info())
            
            _tick_queue.remove(item)

 
def game_thread(callback, *args, **kwargs):
    """
    You can use this function to append a callback to the queue to be executed
    on the next gameframe.
    
    This should be used when you deal with any game-engine related methods in
    your callback as these are only 'really' safe from within the game frame.
    
    @callback  - method to call
    @args      - positional arguments to pass to the callback
    @kwargs    - keyword arguments to pass to the callback
    """
    _queue.put({'cb': callback, 'args': args, 'kwargs': kwargs})


def tick_delay(callback, ticks, *args, **kwargs):
    """
    Delays the execution of a command by a specific amount of ticks. This has
    common use if something has to be delayed for a few ticks to give the
    engine some time to process the information, but is not suitable for a
    longer, threaded delay.
    
    Use this method only when you try to delay something for very few ticks,
    the larger the queue is, the significantly higher the cpu load will be.
    
    Also keep in mind this is not useful for delaying things for a specific
    time as 100 ticks will equal about 3 seconds on a 33 tick server and 1 sec 
    on a 100 tick server.
    
    @callback  - method to call
    @ticks     - amount of ticks to delay the execution of the callback
    @args      - positional arguments to pass to the callback
    @kwargs    - keyword arguments to pass to the callback
    """
    _tick_queue.append({
        'end': _tick + ticks, 
        'cb': callback, 
        'args': args, 
        'kwargs': kwargs,
    })

# =============================================================================
# >> Init
# =============================================================================

_tick = 0 
_queue = Queue()
_tick_queue = []
TickListenerManager.register_listener(_tick_listener)
